/*
#############################################################################

  Ch15p1_SimpleClouds.cpp: a program that demonstrates the water algorithm,
  without any annoying bells and/or whistles.
  
#############################################################################
*/

// include files ////////////////////////////////////////////////////////////
#define STRICT
#include <vector>
#include <stdio.h>
#include <math.h>
#include <D3DX8.h>
#include "D3DApp.h"
#include "D3DFile.h"
#include "D3DFont.h"
#include "D3DUtil.h"
#include "DXUtil.h"
#include "D3DHelperFuncs.h"
#include "Ch15p1_resource.h"
#include "CommonFuncs.h"

using namespace std;
const int NUMOCTAVES = 4;

//===========================================================================
//
// class: CCloudOctave.  Represents a single octave of Perlin noise for our
// cloud.
//
//===========================================================================
class CCloudOctave
{
public:
  CCloudOctave(int iOctaveSize, LPDIRECT3DDEVICE8 pDev,
    LPDIRECT3DVERTEXBUFFER8 pvbQuad);

  virtual ~CCloudOctave();
  
  LPDIRECT3DTEXTURE8 m_ptexOctave; // each octave of noise
  LPDIRECT3DTEXTURE8 m_ptexSmoothOctave; // smoothed octave 

  HRESULT GenerateRegularNoise();
  HRESULT SmoothLayer();

  int m_iOctaveSize;
private:
  LPDIRECT3DDEVICE8 m_pd3dDevice;
  LPDIRECT3DVERTEXBUFFER8 m_pvbQuad;

};

typedef vector<CCloudOctave *> CloudOctaveVector;

CCloudOctave::CCloudOctave(int iOctaveSize, LPDIRECT3DDEVICE8 pDev,
                           LPDIRECT3DVERTEXBUFFER8 pvbQuad)
{
  HRESULT hr;
  // squirrel away the device and vertex buffer pointers for later
  m_pd3dDevice = pDev;
  m_pvbQuad = pvbQuad;
  m_iOctaveSize = iOctaveSize;

  OutputDebugString("\nCCloudOctave::CCloudOctave: Allocating Cloud Octave Textures...");

  if (FAILED(hr = D3DXCreateTexture(pDev, iOctaveSize, iOctaveSize, 
      1, D3DUSAGE_RENDERTARGET, D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, 
      &m_ptexSmoothOctave))) throw(hr);

  if (FAILED(hr = D3DXCreateTexture(pDev, iOctaveSize, iOctaveSize, 
      1, 0, D3DFMT_A8R8G8B8, D3DPOOL_MANAGED, &m_ptexOctave))) throw(hr);
}

CCloudOctave::~CCloudOctave()
{
  OutputDebugString("\nCCloudOctave::CCloudOctave: Freeing Cloud Octave Textures...");
  // the device and vertex buffers do not belong to us, so we don't 
  // delete them.
  SAFE_RELEASE(m_ptexSmoothOctave);
  SAFE_RELEASE(m_ptexOctave);
}

HRESULT CCloudOctave::SmoothLayer()
{
  return(SmoothRenderCopy(m_ptexOctave, m_ptexSmoothOctave, 
    m_iOctaveSize, m_iOctaveSize, m_pd3dDevice, m_pvbQuad));
}

/****************************************************************************

 GenerateRegularNoise: Creates what Kim Pallister refers to as "Salt and
 Pepper" noise.  Each texel is assigned a random shade of gray.

 ****************************************************************************/
HRESULT CCloudOctave::GenerateRegularNoise()
{
  HRESULT hr;
	
	// lock the textures 
  D3DLOCKED_RECT lockedrect;
  ::ZeroMemory(&lockedrect, sizeof(lockedrect));
  
  if (FAILED(hr = m_ptexOctave->LockRect(0, &lockedrect, NULL, 0))) return(hr);
  
  // our texture surface is now locked, and we can use the pitch to traverse it.
  unsigned char *pSurfBits = static_cast<unsigned char *>(lockedrect.pBits);
  
  int index=0;
  for (int y=0; y < m_iOctaveSize; y++) {
    for (int x=0; x < m_iOctaveSize; x++) {
      /*
      
      // Unremark this (and remark out the section below this) to generate 
      // color clouds instead of normal clouds.

      char cIntensityR = (char)((1.0f+Noise3D(x,y,0)) * 128);
      char cIntensityG = (char)((1.0f+Noise3D(x,y,1)) * 128);
      char cIntensityB = (char)((1.0f+Noise3D(x,y,2)) * 128);

      // the intensity determines the color of this texel
      pSurfBits[index++] = cIntensityB; // blue
      pSurfBits[index++] = cIntensityG; // green
      pSurfBits[index++] = cIntensityR; // red
      pSurfBits[index++] = 0xff;       // alpha

      */

      // this chunk of code generates normal clouds

      char cIntensity = (char)((1.0f+Noise2D(x,y)) * 128);
      // the intensity determines the color of this texel
      pSurfBits[index++] = cIntensity; // blue
      pSurfBits[index++] = cIntensity; // green
      pSurfBits[index++] = cIntensity; // red
      pSurfBits[index++] = 0xff;       // alpha
    }
    // next line
    index += lockedrect.Pitch - (m_iOctaveSize*4);
  }

  // unlock texture surface
  if (FAILED(hr = m_ptexOctave->UnlockRect(0))) return(hr);
	return S_OK;
}

//===========================================================================
//
// class: CCloud.  This class represents a cloud texture.  
//
//===========================================================================
class CCloud
{
public:
  // pointers must be set to NULL so that SAFE_RELEASE doesn't try
  // to release something that isn't there.
  CCloud() : m_ptexFinal(NULL), m_ptexComposite(NULL), m_pvbAlphaBlendingQuad(NULL),
    m_iTextureSize(0) { }
  virtual ~CCloud() { InvalidateDeviceObjects(); }

  HRESULT RestoreDeviceObjects(LPDIRECT3DDEVICE8 pDev, 
    LPDIRECT3DVERTEXBUFFER8 pvbQuad, int iNumOctaves, int iTextureSize);
  HRESULT InvalidateDeviceObjects();

  void RenderToFinalTexture();

  // you can change this to m_ptexComposite to see what it looks like
  // without squaring.
  LPDIRECT3DTEXTURE8 GetFinalTexture() { return(m_ptexComposite); }

  CloudOctaveVector m_vOctaves; // individual cloud layers
private:
  
  LPDIRECT3DTEXTURE8 m_ptexComposite; // the composite noise texture
  LPDIRECT3DTEXTURE8 m_ptexFinal; // the final cloud texture (composite squared)

  void CompositeLayers();
  void SquareCompositeLayer();

  LPDIRECT3DVERTEXBUFFER8 m_pvbAlphaBlendingQuad; // 0.5 alpha quad
  
  int m_iTextureSize;

  LPDIRECT3DDEVICE8 m_pd3dDevice;
  LPDIRECT3DVERTEXBUFFER8 m_pvbQuad; // normal (opaque) quad for rendering textures
};

void CCloud::CompositeLayers()
{
  // get surface interfaces
  LPDIRECT3DSURFACE8 psurfComposite;
  m_ptexComposite->GetSurfaceLevel(0, &psurfComposite);

  // set new rendering target & clear
  m_pd3dDevice->SetRenderTarget(psurfComposite, NULL);
  m_pd3dDevice->Clear( 0L, NULL, D3DCLEAR_TARGET, 0x000000, 1.0f, 0L );
  
  // set up orthogonal projection matrix
  SetupOrthoProjForRenderCopy(m_pd3dDevice, m_iTextureSize, m_iTextureSize);

  // begin rendering the scene
  if (FAILED(m_pd3dDevice->BeginScene())) return;
  m_pd3dDevice->SetStreamSource(0, m_pvbAlphaBlendingQuad, sizeof(VERTEX_XYZ_DIFFUSE_TEX1));
  m_pd3dDevice->SetVertexShader(D3DFVF_XYZ_DIFFUSE_TEX1);

  // now alpha blend (using a factor of 0.5) all other octaves onto the final dest.
  // D3D gets the alpha value from our vertices' alpha components.
  
  // set up texture stages for simple texture stage copy
	m_pd3dDevice->SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_MODULATE);
	m_pd3dDevice->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
  m_pd3dDevice->SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
  m_pd3dDevice->SetTextureStageState(1, D3DTSS_COLOROP,   D3DTOP_DISABLE); 

  m_pd3dDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
  m_pd3dDevice->SetRenderState(D3DRS_ALPHATESTENABLE, FALSE);
  m_pd3dDevice->SetRenderState(D3DRS_ZENABLE, D3DZB_FALSE);
  m_pd3dDevice->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_ONE);
	m_pd3dDevice->SetRenderState(D3DRS_DESTBLEND,D3DBLEND_SRCCOLOR);
		
  
  for (CloudOctaveVector::iterator i = m_vOctaves.begin(); i != m_vOctaves.end(); ++i) {
    CCloudOctave *octave = (*i);
  
    // set the source texture active
    m_pd3dDevice->SetTexture(0, octave->m_ptexSmoothOctave);

    // this "blits" the texture
    m_pd3dDevice->DrawPrimitive( D3DPT_TRIANGLELIST, 0, 2 );
  }

  // end scene
  m_pd3dDevice->EndScene();

  // release the dest surface 
  psurfComposite->Release();
}

void CCloud::SquareCompositeLayer()
{
  // get surface interfaces
  LPDIRECT3DSURFACE8 psurfFinal;
  m_ptexFinal->GetSurfaceLevel(0, &psurfFinal);

  // set new rendering target & clear
  m_pd3dDevice->SetRenderTarget(psurfFinal, NULL);
  m_pd3dDevice->Clear( 0L, NULL, D3DCLEAR_TARGET, 0x000000, 1.0f, 0L );
  
  // set up orthogonal projection matrix
  SetupOrthoProjForRenderCopy(m_pd3dDevice, m_iTextureSize, m_iTextureSize);

  // begin rendering the scene
  if (FAILED(m_pd3dDevice->BeginScene())) return;
  
  m_pd3dDevice->SetStreamSource(0, m_pvbQuad, sizeof(VERTEX_XYZ_DIFFUSE_TEX1));
  m_pd3dDevice->SetVertexShader(D3DFVF_XYZ_DIFFUSE_TEX1);
  
  m_pd3dDevice->SetTexture(0,m_ptexComposite);
	m_pd3dDevice->SetTextureStageState(0,D3DTSS_COLOROP, D3DTOP_SELECTARG1);
	m_pd3dDevice->SetTextureStageState(0,D3DTSS_COLORARG1, D3DTA_TEXTURE);
	m_pd3dDevice->SetTextureStageState(0,D3DTSS_COLORARG2, D3DTA_DIFFUSE);
  
  m_pd3dDevice->SetTexture(1,m_ptexComposite);
	m_pd3dDevice->SetTextureStageState(1,D3DTSS_COLOROP, D3DTOP_SELECTARG1);
	m_pd3dDevice->SetTextureStageState(1,D3DTSS_COLORARG1, D3DTA_TEXTURE);
	m_pd3dDevice->SetTextureStageState(1,D3DTSS_COLORARG2, D3DTA_CURRENT);
  
	m_pd3dDevice->SetTexture(2,NULL);
	m_pd3dDevice->SetTextureStageState(2,D3DTSS_COLOROP, D3DTOP_DISABLE); 
	
	m_pd3dDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);

	// this "blits" the texture
  m_pd3dDevice->DrawPrimitive( D3DPT_TRIANGLELIST, 0, 2 );
  
  // end scene
  m_pd3dDevice->EndScene();

  // release the dest surface 
  psurfFinal->Release();
}

void CCloud::RenderToFinalTexture()
{
  // get the current depth buffer (we have to pass this into SetRenderTarget
  // so we don't inadvertently drop our depth buffer.)
  LPDIRECT3DSURFACE8 pDepthSurf;
  m_pd3dDevice->GetDepthStencilSurface(&pDepthSurf);

  // get the current rendering target (we'll set it back later)
  LPDIRECT3DSURFACE8 pOldRenderTarget;
  m_pd3dDevice->GetRenderTarget(&pOldRenderTarget);

  // combine all the noise layers together onto m_ptexComposite
  CompositeLayers();

  // now square the composite texture and put it on the final texture.
  //SquareCompositeLayer();

  // set the rendering target back to the old target.
  m_pd3dDevice->SetRenderTarget(pOldRenderTarget, pDepthSurf);
  m_pd3dDevice->SetRenderState(D3DRS_ZENABLE, D3DZB_TRUE);
  pOldRenderTarget->Release();

  // release the depth surface interface.
  pDepthSurf->Release();

}

HRESULT CCloud::RestoreDeviceObjects(LPDIRECT3DDEVICE8 pDev, 
                                     LPDIRECT3DVERTEXBUFFER8 pvbQuad,
                                     int iNumOctaves, int iTextureSize)
{
  HRESULT hr;
  InvalidateDeviceObjects(); // just in case something's left over

  // squirrel away the device and vertex buffer pointers for later
  m_pd3dDevice = pDev;
  m_pvbQuad = pvbQuad;

  // create our blending vertex buffer... a quad with
  // 0.5 alpha components on all vertices.
  // Create vertex buffer
  if (FAILED(hr = CreateQuad(&m_pvbAlphaBlendingQuad, D3DPOOL_MANAGED, 2.0f, 
    D3DCOLOR_ARGB(128,128,128,128), m_pd3dDevice))) return hr;

  // new up the appropriate number of cloud layers
  int iThisSize = GetLowestPowerOf2(iTextureSize);
  
  m_iTextureSize = iThisSize;

  try {
    for (int loop=0; loop < iNumOctaves; loop++) {
      CCloudOctave *octave = new CCloudOctave(iThisSize, pDev, pvbQuad);
      m_vOctaves.push_back(octave);
      iThisSize /= 2;
    }

    // create the final and composite texture
    if (FAILED(hr = D3DXCreateTexture(pDev, iTextureSize, iTextureSize, 
      1, D3DUSAGE_RENDERTARGET, D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, &m_ptexFinal))) throw(hr);

    if (FAILED(hr = D3DXCreateTexture(pDev, iTextureSize, iTextureSize, 
      1, D3DUSAGE_RENDERTARGET, D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, &m_ptexComposite))) throw(hr);

  } catch(HRESULT hr) { return hr; }

  // now that the layers are constructed, generate some noise for each.
  for (CloudOctaveVector::iterator i = m_vOctaves.begin(); i != m_vOctaves.end(); ++i) {
    if (FAILED(hr = (*i)->GenerateRegularNoise())) return(hr);

    // use bilinear filtering to smooth out the generated noise.
    if (FAILED(hr = (*i)->SmoothLayer())) return(hr);
  }
  return S_OK;
}

HRESULT CCloud::InvalidateDeviceObjects()
{
  // delete the cloud layers
  for (CloudOctaveVector::iterator i = m_vOctaves.begin(); i != m_vOctaves.end(); ++i) {
    delete (*i);
  }
  m_vOctaves.clear();
  
  // release textures
  SAFE_RELEASE(m_ptexFinal);
  SAFE_RELEASE(m_ptexComposite);

  // release alpha blending quad
  SAFE_RELEASE(m_pvbAlphaBlendingQuad);

  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: class CMyD3DApplication
// Desc: Application class. The base class (CD3DApplication) provides the 
//       generic functionality needed in all Direct3D samples. CMyD3DApplication 
//       adds functionality specific to this sample program.
//-----------------------------------------------------------------------------
class CMyD3DApplication : public CD3DApplication
{
  // Font for drawing text
  CD3DFont* m_pFont;
  CD3DFont* m_pFontSmall;

  // quad for rendering our textures
  LPDIRECT3DVERTEXBUFFER8 m_pVB;
  DWORD  m_dwNumVertices;
  
  // Our Cloud
  CCloud m_Cloud;
  
protected:
  HRESULT OneTimeSceneInit();
  HRESULT InitDeviceObjects();
  HRESULT RestoreDeviceObjects();
  HRESULT InvalidateDeviceObjects();
  HRESULT DeleteDeviceObjects();
  HRESULT FinalCleanup();
  HRESULT Render();
  HRESULT FrameMove();
  HRESULT ConfirmDevice( D3DCAPS8* pCaps, DWORD dwBehavior, D3DFORMAT Format );
  LRESULT MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam );

public:
  CMyD3DApplication();
};

//-----------------------------------------------------------------------------
// Name: WinMain()
// Desc: Entry point to the program. Initializes everything, and goes into a
//       message-processing loop. Idle time is used to render the scene.
//-----------------------------------------------------------------------------
INT WINAPI WinMain( HINSTANCE hInst, HINSTANCE, LPSTR, INT )
{
  CMyD3DApplication d3dApp;

  if( FAILED( d3dApp.Create( hInst ) ) )
    return 0;

  return d3dApp.Run();
}

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
CMyD3DApplication::CMyD3DApplication()
{
  m_strWindowTitle    = _T("Ch15p1_SimpleClouds");
  m_bUseDepthBuffer   = TRUE;

  m_pFont            = new CD3DFont( _T("Arial"), 12, D3DFONT_BOLD );
  m_pFontSmall       = new CD3DFont( _T("Arial"),  9, D3DFONT_BOLD );
  m_pVB              = NULL;
  m_dwNumVertices    = 6;
}

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::OneTimeSceneInit()
{
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FrameMove()
{
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::Render()
{
  // Clear the backbuffer
  m_pd3dDevice->Clear(0L, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER,
                      0x000000, 1.0f, 0L);
  
  // Begin rendering the scene
  if( SUCCEEDED(m_pd3dDevice->BeginScene()))
  {
    // set up an orthogonal projection of 1050x1050 units.
    D3DXMATRIX projmat;
    D3DXMatrixOrthoLH(&projmat, 1000.0f, 1000.0f, 0.0, 100.0);
    m_pd3dDevice->SetTransform( D3DTS_PROJECTION, &projmat );

    // draw all of the cloud octaves individually across the top of the screen.
    CloudOctaveVector &vOctaves = m_Cloud.m_vOctaves;
    
    m_pd3dDevice->SetStreamSource(0, m_pVB, sizeof(VERTEX_XYZ_DIFFUSE_TEX1));
    m_pd3dDevice->SetVertexShader(D3DFVF_XYZ_DIFFUSE_TEX1);
    
    m_pd3dDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
    m_pd3dDevice->SetRenderState(D3DRS_ALPHATESTENABLE, FALSE);
    m_pd3dDevice->SetRenderState(D3DRS_ZENABLE, D3DZB_FALSE);

    // set up texture stages for simple texture stage copy
	  m_pd3dDevice->SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_MODULATE);
	  m_pd3dDevice->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
    m_pd3dDevice->SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
    m_pd3dDevice->SetTextureStageState(1, D3DTSS_COLOROP,   D3DTOP_DISABLE); 

    for (int q=0; q < vOctaves.size(); q++) {
      CCloudOctave *oct = vOctaves[q];

      int fOctaveWidth = 900.0f/(float)vOctaves.size();
      int fOctavePos = -450+(q*fOctaveWidth)+(fOctaveWidth/2.0f);
      
      // set up our world transform for this quad
      D3DXMATRIX transmat, scalemat,worldmat;
      D3DXMatrixScaling(&scalemat, (fOctaveWidth-50)/2, (fOctaveWidth-50)/2, 1.0f);
      D3DXMatrixMultiply(&worldmat, &scalemat, D3DXMatrixTranslation(&transmat, fOctavePos, 400.0f, 0.0));
      m_pd3dDevice->SetTransform( D3DTS_WORLD, &worldmat );

      // make the octave's texture active...
      m_pd3dDevice->SetTexture(0, oct->m_ptexSmoothOctave);

      // draw our quad
      m_pd3dDevice->DrawPrimitive(D3DPT_TRIANGLELIST, 0, 2);
    }

    // now we draw the final composite cloud texture in a big square
    D3DXMATRIX transmat, scalemat,worldmat;
    D3DXMatrixScaling(&scalemat, 350.0f, 350.0f, 1.0f);
    D3DXMatrixMultiply(&worldmat, &scalemat, D3DXMatrixTranslation(&transmat, 0.0f, -100.0f, 0.0f));
    m_pd3dDevice->SetTransform( D3DTS_WORLD, &worldmat );


    // make the final cloud texture active...
    m_pd3dDevice->SetTexture(0, m_Cloud.GetFinalTexture());

    // draw our quad
    m_pd3dDevice->DrawPrimitive(D3DPT_TRIANGLELIST, 0, 2);

    // Output statistics
    m_pFont->DrawText(2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats);
    m_pFont->DrawText(2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats);

    // End the scene.
    m_pd3dDevice->EndScene();
  }

  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InitDeviceObjects()
{
  m_pFont->InitDeviceObjects( m_pd3dDevice );
  m_pFontSmall->InitDeviceObjects( m_pd3dDevice );
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::RestoreDeviceObjects()
{
  HRESULT hr;

  m_pFont->RestoreDeviceObjects();
  m_pFontSmall->RestoreDeviceObjects();

  // Create vertex buffer
  if (FAILED( hr = CreateQuad(&m_pVB, D3DPOOL_MANAGED, 2.0f, 0xffffffff, m_pd3dDevice)))
    return hr;
  
  // init the cloud - this sets up all our textures and everything
  if (FAILED(hr = m_Cloud.RestoreDeviceObjects(m_pd3dDevice, m_pVB, NUMOCTAVES, 256)))
    return(hr);

  // since we're not animating, we can tell the cloud to render itself to
  // its texture now.
  m_Cloud.RenderToFinalTexture();
  
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc:
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InvalidateDeviceObjects()
{
  m_pFont->InvalidateDeviceObjects();
  m_pFontSmall->InvalidateDeviceObjects();
  m_Cloud.InvalidateDeviceObjects();

  SAFE_RELEASE(m_pVB);
  
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::DeleteDeviceObjects()
{
  m_pFont->DeleteDeviceObjects();
  m_pFontSmall->DeleteDeviceObjects();
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FinalCleanup()
{
  SAFE_DELETE( m_pFont );
  SAFE_DELETE( m_pFontSmall );
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device intialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::ConfirmDevice( D3DCAPS8* pCaps, DWORD dwBehavior,
                                          D3DFORMAT Format )
{
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Message proc function to handle key and menu input
//-----------------------------------------------------------------------------
LRESULT CMyD3DApplication::MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam,
                                    LPARAM lParam )
{
  return CD3DApplication::MsgProc( hWnd, uMsg, wParam, lParam );
}