/*
#############################################################################

  Ch22p2_InSceneLensFlare.cpp: a program that demonstrates how to create an
  in-scene lens flare effect.
  
#############################################################################
*/

// include files ////////////////////////////////////////////////////////////
#define STRICT
#include <stdio.h>
#include <math.h>
#include <D3DX8.h>
#include "D3DApp.h"
#include "D3DFile.h"
#include "D3DFont.h"
#include "D3DUtil.h"
#include "DXUtil.h"
#include "D3DHelperFuncs.h"
#include "Ch22p2_resource.h"
#include "Camera.h"
#include "InputManager.h"
#include "AnimSequence.h"
#include "Sprite.h"
#include "SkyBox.h"
#include "Ch22p1_LensFlare.h"

const int NUMSPRITES = 5;

//-----------------------------------------------------------------------------
// Name: class CMyD3DApplication
// Desc: Application class. The base class (CD3DApplication) provides the 
//       generic functionality needed in all Direct3D samples. CMyD3DApplication 
//       adds functionality specific to this sample program.
//-----------------------------------------------------------------------------
class CMyD3DApplication : public CD3DApplication
{
  // Font for drawing text
  CD3DFont* m_pFont;
  CD3DFont* m_pFontSmall;

  // Scene
  CUserControlledCamera m_Camera;

  // Mouse Input
  CInputManager m_InputManager;

  // object mesh
  CD3DMesh* m_pObject;
	
  // skybox
  CSkyBox m_SkyBox;

	// lens flare
	CLensFlare m_LensFlare;
	D3DXVECTOR3 m_vSunPos;

	float m_fPlanetRotation;
	int m_iNumObjectsAtPoint;
	int m_iSunScreenPosX;
	int m_iSunScreenPosY;

	D3DXMATRIX m_matProj;

protected:
  HRESULT OneTimeSceneInit();
  HRESULT InitDeviceObjects();
  HRESULT RestoreDeviceObjects();
  HRESULT InvalidateDeviceObjects();
  HRESULT DeleteDeviceObjects();
  HRESULT FinalCleanup();
  HRESULT Render();
  HRESULT FrameMove();
  HRESULT ConfirmDevice( D3DCAPS8* pCaps, DWORD dwBehavior, D3DFORMAT Format );
  LRESULT MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam );

	bool IsLensFlareObscured();
	int NumObjectsAtPoint(int iX, int iY);
  void ProcessInput();

public:
  CMyD3DApplication();
};



//-----------------------------------------------------------------------------
// Name: WinMain()
// Desc: Entry point to the program. Initializes everything, and goes into a
//       message-processing loop. Idle time is used to render the scene.
//-----------------------------------------------------------------------------
INT WINAPI WinMain( HINSTANCE hInst, HINSTANCE, LPSTR, INT )
{
  CMyD3DApplication d3dApp;

  if( FAILED( d3dApp.Create( hInst ) ) )
    return 0;

  return d3dApp.Run();
}

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
CMyD3DApplication::CMyD3DApplication()
{
  m_strWindowTitle    = _T("Ch22p2_InSceneLensFlare");
  m_bUseDepthBuffer   = TRUE;
  m_pObject           = new CD3DMesh;
  m_pFont            = new CD3DFont( _T("Arial"), 12, D3DFONT_BOLD );
  m_pFontSmall       = new CD3DFont( _T("Arial"),  9, D3DFONT_BOLD );
}

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::OneTimeSceneInit()
{
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FrameMove()
{
	if (m_fElapsedTime < 0.0001f) {
		m_fElapsedTime = 0.0001f;
	}
	if (m_fElapsedTime > 0.5f) m_fElapsedTime = 0.5f;

	char buf[256]; sprintf(buf, "\nElapsed Time: %f", m_fElapsedTime);
	OutputDebugString(buf);

  CTimer::UpdateAll(m_fElapsedTime);
	m_fPlanetRotation += m_fElapsedTime;
	if (m_fPlanetRotation > (float)(32.0f*PI)) m_fPlanetRotation -= (float)(32.0f*PI);
  return S_OK;
}

void CMyD3DApplication::ProcessInput()
{
  const float fSpeed = 0.5f;
  unsigned char m_bKey[256];
  ZeroMemory( m_bKey, 256 );
  GetKeyboardState(m_bKey);
  // Process keyboard input
  if(m_bKey['D'] & 128) m_Camera.AddToVelocity(D3DXVECTOR3(fSpeed, 0.0f, 0.0f)); // Slide Right
  if(m_bKey['A'] & 128) m_Camera.AddToVelocity(D3DXVECTOR3(-fSpeed, 0.0f, 0.0f));// Slide Left
  if(m_bKey['Q'] & 128) m_Camera.AddToVelocity(D3DXVECTOR3(0.0f, fSpeed, 0.0f)); // Slide Up
  if(m_bKey['Z'] & 128) m_Camera.AddToVelocity(D3DXVECTOR3(0.0f, -fSpeed, 0.0f));// Slide Down
  if(m_bKey['W'] & 128) m_Camera.AddToVelocity(D3DXVECTOR3(0.0f, 0.0f, fSpeed)); // Slide Foward
  if(m_bKey['S'] & 128) m_Camera.AddToVelocity(D3DXVECTOR3(0.0f, 0.0f, -fSpeed));// Slide Back
  if(m_bKey['L'] & 128) m_Camera.AddToYawPitchRoll(fSpeed, 0.0f, 0.0f);  // Turn Right
  if(m_bKey['J'] & 128) m_Camera.AddToYawPitchRoll(-fSpeed, 0.0f, 0.0f); // Turn Left
  if(m_bKey['K'] & 128) m_Camera.AddToYawPitchRoll(0.0f, fSpeed, 0.0f); // Turn Down
  if(m_bKey['I'] & 128) m_Camera.AddToYawPitchRoll(0.0f, -fSpeed, 0.0f);// Turn Up

  // mouse look
  DIMOUSESTATE2 dims2;
  m_InputManager.ReadMouse(dims2);

  // play with the divisor constants to change the mouselook sensitivity.
  // I've found that these values most accurately simulate my beloved Q3A setup. :)
  m_Camera.AddToYawPitchRoll((float)dims2.lX/0.8f, (float)dims2.lY/0.8f, 0.0f);
  
}

bool CMyD3DApplication::IsLensFlareObscured()
{
	// get screen coordinates of the light source.
	int iLightW;
	
	m_LensFlare.CalcLightSourceScreenCoords(m_Camera, m_matProj, m_vSunPos, 
		m_d3dsdBackBuffer.Width,m_d3dsdBackBuffer.Height, m_iSunScreenPosX, 
		m_iSunScreenPosY, iLightW);

	m_iNumObjectsAtPoint = NumObjectsAtPoint(m_iSunScreenPosX, m_iSunScreenPosY);
	return(m_iNumObjectsAtPoint > 0);
	
}

int CMyD3DApplication::NumObjectsAtPoint(int iX, int iY)
{
	D3DXVECTOR3 vRayDir;
  D3DXVECTOR3 vRayOrig;

  int iNumObjectsAtPoint = 0;

  // Get the pick ray from given coordinates
	{
		D3DXVECTOR3 v;
		v.x =  ( ( ( 2.0f * iX ) / m_d3dsdBackBuffer.Width  ) - 1 ) / m_matProj._11;
		v.y = -( ( ( 2.0f * iY ) / m_d3dsdBackBuffer.Height ) - 1 ) / m_matProj._22;
		v.z =  1.0f;

		// Get the inverse view matrix
		D3DXMATRIX matView, m;
		matView = m_Camera.GetViewMatrix();
		D3DXMatrixInverse( &m, NULL, &matView );

		// Transform the screen space pick ray into 3D space
		vRayDir.x  = v.x*m._11 + v.y*m._21 + v.z*m._31;
		vRayDir.y  = v.x*m._12 + v.y*m._22 + v.z*m._32;
		vRayDir.z  = v.x*m._13 + v.y*m._23 + v.z*m._33;
		vRayOrig.x = m._41;
		vRayOrig.y = m._42;
		vRayOrig.z = m._43;
	}
  
  // Get the picked triangle
  LPDIRECT3DVERTEXBUFFER8 pVB;
  LPDIRECT3DINDEXBUFFER8 pIB;
  m_pObject->GetLocalMesh()->GetVertexBuffer( &pVB );
  m_pObject->GetLocalMesh()->GetIndexBuffer( &pIB );

  struct VERTEX { D3DXVECTOR3 p, n; FLOAT tu; FLOAT tv; };
  DWORD dwFVF = m_pObject->GetLocalMesh()->GetFVF();

  WORD*      pIndices;
  VERTEX*    pVertices;
  DWORD      dwNumFaces = m_pObject->GetLocalMesh()->GetNumFaces();
  pIB->Lock( 0,0,(BYTE**)&pIndices, 0 );
  pVB->Lock( 0,0,(BYTE**)&pVertices, 0 );

  for( DWORD i=0; i<dwNumFaces; i++ )
  {
      D3DXVECTOR3 v0 = pVertices[pIndices[3*i+0]].p;
      D3DXVECTOR3 v1 = pVertices[pIndices[3*i+1]].p;
      D3DXVECTOR3 v2 = pVertices[pIndices[3*i+2]].p;

			float fPickT, fPickU, fPickV;
      // Check if the pick ray passes through this point
      if( DoesRayIntersectTriangle( vRayOrig, vRayDir, v0, v1, v2,
                             &fPickT, &fPickU, &fPickV ) )
      {
          iNumObjectsAtPoint++;
      }
  }

  pVB->Unlock();
  pIB->Unlock();

  pVB->Release();
  pIB->Release();
  return(iNumObjectsAtPoint);
}



//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::Render()
{
  
  // these are done here so that you can move the camera around during a freeze
  // frame, ala The Matrix
  D3DXMATRIX matWorld;
  D3DXMatrixIdentity( &matWorld );
	m_pd3dDevice->SetTransform( D3DTS_WORLD, &matWorld );

  ProcessInput();
  m_Camera.Update(m_fElapsedTime);

  m_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
  m_pd3dDevice->SetRenderState( D3DRS_SRCBLEND, D3DBLEND_SRCALPHA );
  m_pd3dDevice->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA );
  
  // Clear the backbuffer
  m_pd3dDevice->Clear( 0L, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER,
                       0x000000, 1.0f, 0L );
  
  if( SUCCEEDED( m_pd3dDevice->BeginScene() ) )
  {
    // draw skybox
		m_pd3dDevice->SetRenderState(D3DRS_ZENABLE, FALSE);
		m_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, FALSE);
    m_SkyBox.Render(m_Camera.GetViewMatrix());

		m_pd3dDevice->SetRenderState(D3DRS_ZENABLE, TRUE);
		
		m_pd3dDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
		m_pd3dDevice->SetTransform( D3DTS_VIEW, &m_Camera.GetViewMatrix());
    m_pd3dDevice->SetVertexShader( D3DFVF_XYZ_DIFFUSE_TEX1 );

		// draw Object

		bool bLensFlareObscured = IsLensFlareObscured();
		if (bLensFlareObscured) {
			// the lens flare is obscured, so draw the only halo first then draw
			// the planet.
			m_pd3dDevice->SetRenderState(D3DRS_ZENABLE, FALSE);

  		m_LensFlare.Render(m_Camera, m_matProj, m_vSunPos, 
	  		m_d3dsdBackBuffer.Width, m_d3dsdBackBuffer.Height, true);
		}

		m_pd3dDevice->SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);
		m_pd3dDevice->SetTextureStageState(0, D3DTSS_ALPHAOP,   D3DTOP_SELECTARG1);

		m_pd3dDevice->SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_LINEAR);
		m_pd3dDevice->SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR);


		m_pd3dDevice->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
		m_pd3dDevice->SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
		m_pd3dDevice->SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_MODULATE);


		m_pd3dDevice->SetVertexShader( D3DFVF_XYZ_DIFFUSE_TEX1 );
		m_pd3dDevice->SetTransform( D3DTS_VIEW, &m_Camera.GetViewMatrix());
		m_pd3dDevice->SetTransform( D3DTS_PROJECTION, &m_matProj);
		m_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, FALSE);

    D3DXMatrixRotationYawPitchRoll(&matWorld, -m_fPlanetRotation/16.0f, 0.0f, 0.0f);
    m_pd3dDevice->SetTransform( D3DTS_WORLD, &matWorld );
		m_pd3dDevice->SetRenderState(D3DRS_ZENABLE, TRUE);
    m_pd3dDevice->SetRenderState( D3DRS_LIGHTING, TRUE);
    m_pd3dDevice->SetRenderState(D3DRS_AMBIENT, D3DXCOLOR(0.2f, 0.2f, 0.2f, 1.0f));
    m_pObject->Render( m_pd3dDevice );
    m_pd3dDevice->SetRenderState( D3DRS_LIGHTING, FALSE);
		
		if (!bLensFlareObscured) {
			// the lens flare isn't obscured, so draw it last, and draw the whole 
			// thing
  		m_pd3dDevice->SetRenderState(D3DRS_ZENABLE, false);

		  m_LensFlare.Render(m_Camera, m_matProj, m_vSunPos, 
			  m_d3dsdBackBuffer.Width, m_d3dsdBackBuffer.Height, false);

		}

		m_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, FALSE);
    
    char buf[256];
    _snprintf(buf, 256, "Position: (%.2f, %.2f, %.2f)", 
      (float)m_Camera.GetPosition().x, (float)m_Camera.GetPosition().y, 
      (float)m_Camera.GetPosition().z);

    m_pFontSmall->DrawText( 2,  0,  D3DCOLOR_ARGB(255,255,255,0), buf );
    
		_snprintf(buf, 256, "Screen Pos of Sun: (%d,%d)", m_iSunScreenPosX, m_iSunScreenPosY);

    m_pFontSmall->DrawText( 2,  15,  D3DCOLOR_ARGB(255,255,255,0), buf );
    
		_snprintf(buf, 256, "NumObjectsAtPoint: %d", m_iNumObjectsAtPoint);

    m_pFontSmall->DrawText( 2,  30,  D3DCOLOR_ARGB(255,255,255,0), buf );
    
    // End the scene.
    m_pd3dDevice->EndScene();
  }

  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InitDeviceObjects()
{
  m_pFont->InitDeviceObjects( m_pd3dDevice );
  m_pFontSmall->InitDeviceObjects( m_pd3dDevice );
	m_SkyBox.SetSize(100.0f);
  
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::RestoreDeviceObjects()
{
	m_fPlanetRotation = 0.0f;

	m_vSunPos = D3DXVECTOR3(10, 5, 50);
	
  m_InputManager.CreateDevices(m_hWnd, false, true);

  m_SkyBox.RestoreDeviceObjects(m_pd3dDevice,
    "Ch22p2_SkyBox_Top.bmp",
    "Ch22p2_SkyBox_Bottom.bmp",
    "Ch22p2_SkyBox_Front.bmp",
    "Ch22p2_SkyBox_Back.bmp",
    "Ch22p2_SkyBox_Left.bmp",
    "Ch22p2_SkyBox_Right.bmp"
    );

	m_LensFlare.RestoreDeviceObjects(m_pd3dDevice);

	LPDIRECT3DTEXTURE8 texHalo = m_LensFlare.AddTexture("Ch22p2_LensFlare_Halo.dds");
  LPDIRECT3DTEXTURE8 tex1 = m_LensFlare.AddTexture("Ch22p2_LensFlare_01.dds");
  LPDIRECT3DTEXTURE8 tex2 = m_LensFlare.AddTexture("Ch22p2_LensFlare_02.dds");
  LPDIRECT3DTEXTURE8 tex3 = m_LensFlare.AddTexture("Ch22p2_LensFlare_03.dds");
	
  m_LensFlare.AddSpot(CLensFlareSpot(texHalo, 1.20f,  1.0f, D3DXCOLOR(1.0f, 1.0f, 1.0f, 1.0f)));
  
	m_LensFlare.AddSpot(CLensFlareSpot(tex2, 0.10f,  0.8f, D3DXCOLOR(0.7f, 0.5f, 0.0f, 0.2f)));
  m_LensFlare.AddSpot(CLensFlareSpot(tex2, 0.01f,  0.7f, D3DXCOLOR(1.0f, 0.0f, 0.0f, 0.7f)));
	m_LensFlare.AddSpot(CLensFlareSpot(tex1, 0.05f,  0.6f, D3DXCOLOR(1.0f, 1.0f, 0.0f, 0.5f)));
	m_LensFlare.AddSpot(CLensFlareSpot(tex3, 0.13f,  0.5f, D3DXCOLOR(1.0f, 1.0f, 0.0f, 0.5f)));
	m_LensFlare.AddSpot(CLensFlareSpot(tex1, 0.05f,  0.4f, D3DXCOLOR(1.0f, 0.5f, 0.0f, 1.0f)));
  m_LensFlare.AddSpot(CLensFlareSpot(tex2, 0.05f,  0.1f, D3DXCOLOR(1.0f, 1.0f, 0.5f, 0.5f)));
	m_LensFlare.AddSpot(CLensFlareSpot(tex1, 0.05f, -0.2f, D3DXCOLOR(1.0f, 0.0f, 0.0f, 1.0f)));
  m_LensFlare.AddSpot(CLensFlareSpot(tex2, 0.09f, -0.3f, D3DXCOLOR(1.0f, 1.0f, 0.6f, 0.5f)));
	m_LensFlare.AddSpot(CLensFlareSpot(tex1, 0.15f, -0.4f, D3DXCOLOR(1.0f, 0.7f, 0.0f, 0.3f)));
  m_LensFlare.AddSpot(CLensFlareSpot(tex3, 0.15f, -0.7f, D3DXCOLOR(1.0f, 0.5f, 0.0f, 0.2f)));
	m_LensFlare.AddSpot(CLensFlareSpot(tex2, 0.32f, -1.0f, D3DXCOLOR(1.0f, 0.7f, 0.0f, 0.4f)));
	m_LensFlare.AddSpot(CLensFlareSpot(tex3, 0.40f, -1.3f, D3DXCOLOR(1.0f, 0.0f, 0.0f, 0.7f)));
	
	m_pFont->RestoreDeviceObjects();
  m_pFontSmall->RestoreDeviceObjects();
  m_Camera.SetPosition(D3DXVECTOR3(0.0f, 7.0f, -20.0f));
	
	if( FAILED( m_pObject->Create(m_pd3dDevice, _T("Ch22p2_Planet.x") ) ) )
    return E_FAIL;
	m_pObject->RestoreDeviceObjects(m_pd3dDevice);

  // Set the world matrix
  D3DXMATRIX matWorld;
  D3DXMatrixIdentity( &matWorld );
  m_pd3dDevice->SetTransform( D3DTS_WORLD, &matWorld );

  // Set projection matrix
  FLOAT fAspect = ((FLOAT)m_d3dsdBackBuffer.Width) / m_d3dsdBackBuffer.Height;
  D3DXMatrixPerspectiveFovLH( &m_matProj, D3DX_PI/4, fAspect, 0.1f, 100.0f );
  m_pd3dDevice->SetTransform( D3DTS_PROJECTION, &m_matProj );
	
  m_pd3dDevice->SetRenderState( D3DRS_SRCBLEND,  D3DBLEND_ONE );
  m_pd3dDevice->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_ONE );
  m_pd3dDevice->SetRenderState( D3DRS_LIGHTING,  FALSE );
  m_pd3dDevice->SetRenderState( D3DRS_CULLMODE,  D3DCULL_CCW );
  m_pd3dDevice->SetRenderState( D3DRS_SHADEMODE, D3DSHADE_GOURAUD );
  
  // create a light for our Object
  D3DLIGHT8 light; ::ZeroMemory(&light, sizeof(D3DLIGHT8));
  light.Ambient.a = light.Ambient.g = light.Ambient.b = light.Ambient.r = 0.0;
  light.Diffuse.a = light.Diffuse.g = light.Diffuse.b = light.Diffuse.r = 1.0;
  light.Position = m_vSunPos;
	light.Range = 500.0f;
  light.Type = D3DLIGHT_POINT;
	light.Phi = 0.0;
	light.Attenuation0 = 1.0f;
	//light.Attenuation1 = 1.0f;
	
  m_pd3dDevice->SetRenderState( D3DRS_LIGHTING, TRUE);
  m_pd3dDevice->SetLight(0, &light);
  m_pd3dDevice->LightEnable(0, true);
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc:
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InvalidateDeviceObjects()
{
  m_InputManager.DestroyDevices();
  m_pObject->InvalidateDeviceObjects();
  m_pFont->InvalidateDeviceObjects();
  m_pFontSmall->InvalidateDeviceObjects();
  m_SkyBox.InvalidateDeviceObjects();
	m_LensFlare.InvalidateDeviceObjects();
	return S_OK;
}

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::DeleteDeviceObjects()
{
  m_pFont->DeleteDeviceObjects();
  m_pFontSmall->DeleteDeviceObjects();
  m_pObject->Destroy();
    
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FinalCleanup()
{
  SAFE_DELETE( m_pFont );
  SAFE_DELETE( m_pFontSmall );
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device intialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::ConfirmDevice( D3DCAPS8* pCaps, DWORD dwBehavior,
                                          D3DFORMAT Format )
{
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Message proc function to handle key and menu input
//-----------------------------------------------------------------------------
LRESULT CMyD3DApplication::MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam,
                                    LPARAM lParam )
{
  // Pass remaining messages to default handler
  return CD3DApplication::MsgProc( hWnd, uMsg, wParam, lParam );
}