/*
#############################################################################

  Ch23p1_SimpleWater.cpp: a program that demonstrates the 3D water effect.
  
#############################################################################
*/

// include files ////////////////////////////////////////////////////////////
#define STRICT
#include <stdio.h>
#include <math.h>
#include <D3DX8.h>
#include "D3DApp.h"
#include "D3DFile.h"
#include "D3DFont.h"
#include "D3DUtil.h"
#include "DXUtil.h"
#include "D3DHelperFuncs.h"
#include "Ch23p1_resource.h"
#include "Camera.h"
#include "InputManager.h"
#include "SkyBox.h"
#include "Timer.h"
#include "GroundPlane.h"
#include "Ch23p1_WaterPlane.h"

const int NUMSPRITES = 5;
const int CUBEMAP_RESOLUTION = 256;

//-----------------------------------------------------------------------------
// Name: class CMyD3DApplication
// Desc: Application class. The base class (CD3DApplication) provides the 
//       generic functionality needed in all Direct3D samples. CMyD3DApplication 
//       adds functionality specific to this sample program.
//-----------------------------------------------------------------------------
class CMyD3DApplication : public CD3DApplication
{
  // Font for drawing text
  CD3DFont* m_pFont;
  CD3DFont* m_pFontSmall;

  // Scene
  CUserControlledCamera m_Camera;

  // Mouse Input
  CInputManager m_InputManager;

  // skybox
  CSkyBox m_SkyBox;

	// ground plane
	CGroundPlane m_GroundPlane;

  // pool frame
  CD3DMesh* m_pPoolFrame;
  
  // water
  CWaterPlane m_Water;

  LPDIRECT3DCUBETEXTURE8 m_pCubeMap;
  IDirect3DSurface8* m_pCubeFaceZBuffer;

protected:
  HRESULT OneTimeSceneInit();
  HRESULT InitDeviceObjects();
  HRESULT RestoreDeviceObjects();
  HRESULT InvalidateDeviceObjects();
  HRESULT DeleteDeviceObjects();
  HRESULT FinalCleanup();
  HRESULT Render() { return(ScreenRender()); }
  HRESULT ScreenRender();
  HRESULT CoreRender(bool bRenderWater, D3DXMATRIX matView);
  HRESULT CubeMapRender();
  HRESULT FrameMove();
  HRESULT ConfirmDevice( D3DCAPS8* pCaps, DWORD dwBehavior, D3DFORMAT Format );
  LRESULT MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam );

	void ProcessInput();
	int m_iLightPosX;
	int m_iLightPosY;

public:
  CMyD3DApplication();
};

//-----------------------------------------------------------------------------
// Name: WinMain()
// Desc: Entry point to the program. Initializes everything, and goes into a
//       message-processing loop. Idle time is used to render the scene.
//-----------------------------------------------------------------------------
INT WINAPI WinMain( HINSTANCE hInst, HINSTANCE, LPSTR, INT )
{
  CMyD3DApplication d3dApp;

  if( FAILED( d3dApp.Create( hInst ) ) )
    return 0;

  return d3dApp.Run();
}

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
CMyD3DApplication::CMyD3DApplication()
{
	m_strWindowTitle    = _T("Ch23p1_SimpleWater");
  m_bUseDepthBuffer   = TRUE;
  m_pFont            = new CD3DFont( _T("Arial"), 12, D3DFONT_BOLD );
  m_pFontSmall       = new CD3DFont( _T("Arial"),  9, D3DFONT_BOLD );\
  m_pCubeMap          = NULL;
  m_pCubeFaceZBuffer  = NULL;
}

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::OneTimeSceneInit()
{
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FrameMove()
{
	if (m_fElapsedTime < 0.0001f) {
		m_fElapsedTime = 0.0001f;
	}
	
  ProcessInput();
  m_Camera.Update(m_fElapsedTime);

	CTimer::UpdateAll(m_fElapsedTime);
  m_Water.Update(m_fElapsedTime);

  // Render the scene into the surfaces of the cubemap
  if( FAILED( CubeMapRender() ) )
      return E_FAIL;


	return S_OK;
}

void CMyD3DApplication::ProcessInput()
{
  const float fSpeed = 0.5f;
  unsigned char m_bKey[256];
  ZeroMemory( m_bKey, 256 );
  GetKeyboardState(m_bKey);
  // Process keyboard input
  if(m_bKey['D'] & 128) m_Camera.AddToVelocity(D3DXVECTOR3(fSpeed, 0.0f, 0.0f)); // Slide Right
  if(m_bKey['A'] & 128) m_Camera.AddToVelocity(D3DXVECTOR3(-fSpeed, 0.0f, 0.0f));// Slide Left
  if(m_bKey['Q'] & 128) m_Camera.AddToVelocity(D3DXVECTOR3(0.0f, fSpeed, 0.0f)); // Slide Up
  if(m_bKey['Z'] & 128) m_Camera.AddToVelocity(D3DXVECTOR3(0.0f, -fSpeed, 0.0f));// Slide Down
  if(m_bKey['W'] & 128) m_Camera.AddToVelocity(D3DXVECTOR3(0.0f, 0.0f, fSpeed)); // Slide Foward
  if(m_bKey['S'] & 128) m_Camera.AddToVelocity(D3DXVECTOR3(0.0f, 0.0f, -fSpeed));// Slide Back

  // mouse look
  DIMOUSESTATE2 dims2;
  m_InputManager.ReadMouse(dims2);

  // play with the divisor constants to change the mouselook sensitivity.
  // I've found that these values most accurately simulate my beloved Q3A setup. :)
  m_Camera.AddToYawPitchRoll((float)dims2.lX/0.8f, (float)dims2.lY/0.8f, 0.0f);

	// prevent camera from moving down too far.
	D3DXVECTOR3 pos = m_Camera.GetPosition();
	if (pos.y < 1.7f) pos.y = 1.70f; 
	m_Camera.SetPosition(pos);
}

HRESULT CMyD3DApplication::CubeMapRender()
{
  // Set the projection matrix for a field of view of 90 degrees
  D3DXMATRIX matProj;
  D3DXMatrixPerspectiveFovLH( &matProj, D3DX_PI/2, 1.0f, 0.5f, 100.0f );
  m_pd3dDevice->SetTransform( D3DTS_PROJECTION, &matProj );

  // Get the current view matrix, to concat it with the cubemap view vectors
  D3DXMATRIX matViewDir = m_Camera.GetViewMatrix();
  matViewDir._41 = 0.0f; matViewDir._42 = 0.0f; matViewDir._43 = 0.0f;
  D3DXPLANE plane; plane.a = plane.b = plane.c = plane.d = 0.0f;
  D3DXMATRIX matFlip; D3DXMatrixReflect(&matFlip, &plane);
  
  D3DXMatrixMultiply(&matViewDir, &matFlip, &matViewDir);
  // Store the current backbuffer and zbuffer
  LPDIRECT3DSURFACE8 pBackBuffer, pZBuffer;
  m_pd3dDevice->GetRenderTarget( &pBackBuffer );
  m_pd3dDevice->GetDepthStencilSurface( &pZBuffer );

  // Render to the six faces of the cube map
  for( DWORD i=0; i<6; i++ )
  {
    // Set the view transform for this cubemap surface
    D3DXMATRIX matView;
    matView = D3DUtil_GetCubeMapViewMatrix( (D3DCUBEMAP_FACES)i );
    D3DXMatrixMultiply( &matView, &matViewDir, &matView );
    m_pd3dDevice->SetTransform( D3DTS_VIEW, &matView );

    // Set the rendertarget to the i'th cubemap surface
    LPDIRECT3DSURFACE8 pCubeMapFace;
    m_pCubeMap->GetCubeMapSurface( (D3DCUBEMAP_FACES)i, 0, &pCubeMapFace );
    m_pd3dDevice->SetRenderTarget( pCubeMapFace, m_pCubeFaceZBuffer);
    pCubeMapFace->Release();

    m_pd3dDevice->BeginScene();
    m_pd3dDevice->Clear( 0L, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, 0xffffffff, 1.0f, 0L );
    CoreRender( FALSE, matView );
    m_pd3dDevice->EndScene();
  }

  m_pd3dDevice->SetTransform( D3DTS_VIEW, &m_Camera.GetViewMatrix());
  // Change the rendertarget back to the main backbuffer
  m_pd3dDevice->SetRenderTarget( pBackBuffer, pZBuffer );
  pBackBuffer->Release();
  pZBuffer->Release();

  return S_OK;
}
//-----------------------------------------------------------------------------
// Name: ScreenRender()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::ScreenRender()
{
  if( SUCCEEDED( m_pd3dDevice->BeginScene() ) ) {

    D3DXMATRIX matProj;
    D3DXMatrixPerspectiveFovLH( &matProj, D3DX_PI/4, 1.0f, 0.5f, 100.0f );
    m_pd3dDevice->SetTransform( D3DTS_PROJECTION, &matProj );


    CoreRender(true, m_Camera.GetViewMatrix());

    char buf[256];
    _snprintf(buf, 256, "Position: (%.2f, %.2f, %.2f)", 
      (float)m_Camera.GetPosition().x, (float)m_Camera.GetPosition().y, 
      (float)m_Camera.GetPosition().z);

    m_pFontSmall->DrawText( 2,  0,  D3DCOLOR_ARGB(255,255,255,0), buf );
    
    _snprintf(buf, 256, "Normals are: %s ('N' to switch)", 
      m_Water.GetHardcodeNormals() ? "Hard-coded to (0,1,0)" : "Calculated");

    m_pFontSmall->DrawText( 2,  15,  D3DCOLOR_ARGB(255,255,255,0), buf );

    _snprintf(buf, 256, "Environment Blend Factor: %0.1f (I,O to adjust)", 
      m_Water.GetEnvBlendFactor());

    m_pFontSmall->DrawText( 2,  30,  D3DCOLOR_ARGB(255,255,255,0), buf );
    
    _snprintf(buf, 256, "Simulated Water Depth: %0.1f (K,L to adjust)", 
      m_Water.GetDepth());

    m_pFontSmall->DrawText( 2,  45,  D3DCOLOR_ARGB(255,255,255,0), buf );
    
    _snprintf(buf, 256, "Press C to create a water drop.", 
      m_Water.GetDepth());

    m_pFontSmall->DrawText( 2,  60,  D3DCOLOR_ARGB(255,255,255,0), buf );
    
    // End the scene.
    m_pd3dDevice->EndScene();
  }
  return(S_OK);
}

HRESULT CMyD3DApplication::CoreRender(bool bRenderWater, D3DXMATRIX matView)
{
  D3DXMATRIX matWorld;
  D3DXMatrixIdentity( &matWorld );
  m_pd3dDevice->SetTransform( D3DTS_WORLD, &matWorld );

  // Clear the backbuffer
  m_pd3dDevice->Clear( 0L, NULL, D3DCLEAR_ZBUFFER, 0x000000ff, 1.0f, 0L );
  
	m_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLOROP,   D3DTOP_SELECTARG1 );
	m_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLORARG1, D3DTA_TEXTURE );
	m_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLORARG2, D3DTA_DIFFUSE );
	m_pd3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAOP,   D3DTOP_SELECTARG1 );
	m_pd3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAARG1, D3DTA_DIFFUSE );
	
	m_pd3dDevice->SetTextureStageState( 1, D3DTSS_COLOROP,   D3DTOP_DISABLE );
	m_pd3dDevice->SetTextureStageState( 1, D3DTSS_ALPHAOP,   D3DTOP_DISABLE );

  // draw skybox
	m_pd3dDevice->SetVertexShader(D3DFVF_XYZ_DIFFUSE_TEX1);
	m_pd3dDevice->SetRenderState(D3DRS_ZENABLE, FALSE);
	m_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, FALSE);
	m_pd3dDevice->SetTransform( D3DTS_VIEW, &matView);
  m_pd3dDevice->SetRenderState( D3DRS_LIGHTING, FALSE);
  m_SkyBox.Render(matView);

	// draw ground plane
  D3DXMatrixIdentity( &matWorld );
  m_pd3dDevice->SetTransform( D3DTS_WORLD, &matWorld );
  m_pd3dDevice->SetTransform( D3DTS_VIEW, &matView);
	m_pd3dDevice->SetTexture( 0, m_GroundPlane.GetTexture() );
  m_pd3dDevice->SetStreamSource( 0, m_GroundPlane.GetVB(), sizeof(VERTEX_XYZ_DIFFUSE_TEX1) );
  m_pd3dDevice->SetIndices( m_GroundPlane.GetIB(), 0L );
  m_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, 
    m_GroundPlane.GetNumVerts(), 0, m_GroundPlane.GetNumIndices()/3 );

   // draw water
  D3DLIGHT8 light; ZeroMemory(&light, sizeof(D3DLIGHT8));

	D3DUtil_InitLight(light, D3DLIGHT_DIRECTIONAL, -0.0f, 20.0f, -10.0f);
	light.Ambient.a = 1.0f; 
	light.Ambient.g = 0.4f;
	light.Ambient.b = 0.4f;
	light.Ambient.r = 0.4f;
	light.Diffuse.a = 1.0f; 
	light.Diffuse.g = 0.8f;
	light.Diffuse.b = 0.8f;
	light.Diffuse.r = 0.8f;
	light.Direction = D3DXVECTOR3(0.0f, -1.0f, 0.0f);

	m_pd3dDevice->SetLight(0, &light);
	m_pd3dDevice->LightEnable(0, true);
  m_pd3dDevice->SetRenderState(D3DRS_ZENABLE, TRUE);

  if (bRenderWater) {
    
    D3DXMATRIX matTrans, matScale;
    D3DXMatrixTranslation(&matTrans, 0.0f, 1.1f, 0.0f);
    D3DXMatrixScaling(&matScale, 16.0f, 1.0f, 20.0f);
    matWorld = matScale * matTrans;
		m_pd3dDevice->SetTransform( D3DTS_WORLD, &matWorld );
  
    m_pd3dDevice->SetRenderState( D3DRS_LIGHTING, FALSE);
    m_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, FALSE);
    
    // Turn on texture-coord generation for cubemapping
    m_pd3dDevice->SetTextureStageState( 0, D3DTSS_TEXCOORDINDEX, D3DTSS_TCI_CAMERASPACEREFLECTIONVECTOR );
    m_pd3dDevice->SetTextureStageState( 0, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_COUNT3 );
    m_pd3dDevice->SetTextureStageState( 0, D3DTSS_ADDRESSU,  D3DTADDRESS_MIRROR );
    m_pd3dDevice->SetTextureStageState( 0, D3DTSS_ADDRESSV,  D3DTADDRESS_MIRROR );

    m_pd3dDevice->SetTextureStageState( 1, D3DTSS_COLOROP,   D3DTOP_BLENDDIFFUSEALPHA );
	  m_pd3dDevice->SetTextureStageState( 1, D3DTSS_COLORARG1, D3DTA_CURRENT );
	  m_pd3dDevice->SetTextureStageState( 1, D3DTSS_COLORARG2, D3DTA_TEXTURE );
	  m_pd3dDevice->SetTextureStageState( 1, D3DTSS_ALPHAOP,   D3DTOP_SELECTARG1 );
	  m_pd3dDevice->SetTextureStageState( 1, D3DTSS_ALPHAARG1, D3DTA_TEXTURE );
	  
    m_pd3dDevice->SetTextureStageState( 2, D3DTSS_COLOROP,   D3DTOP_DISABLE );
  	m_pd3dDevice->SetTextureStageState( 2, D3DTSS_ALPHAOP,   D3DTOP_DISABLE );




    m_pd3dDevice->SetTexture( 0, m_pCubeMap );
    
    m_Water.Render();

    // Restore the render states
    m_pd3dDevice->SetTextureStageState( 0, D3DTSS_TEXCOORDINDEX, D3DTSS_TCI_PASSTHRU );
    m_pd3dDevice->SetTextureStageState( 0, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_DISABLE );
  }

  m_pd3dDevice->SetTextureStageState( 1, D3DTSS_COLOROP,   D3DTOP_DISABLE );
	m_pd3dDevice->SetTextureStageState( 1, D3DTSS_ALPHAOP,   D3DTOP_DISABLE );

  D3DXMatrixIdentity(&matWorld);
  m_pd3dDevice->SetTransform( D3DTS_WORLD, &matWorld );
  m_pd3dDevice->SetVertexShader(D3DFVF_XYZ_DIFFUSE_TEX1);
  m_pd3dDevice->SetRenderState( D3DRS_LIGHTING, TRUE);
	
  m_pPoolFrame->Render(m_pd3dDevice);

  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InitDeviceObjects()
{
  m_pFont->InitDeviceObjects( m_pd3dDevice );
  m_pFontSmall->InitDeviceObjects( m_pd3dDevice );
	m_SkyBox.SetSize(100.0f);
  
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::RestoreDeviceObjects()
{
  HRESULT hr;
	m_iLightPosX = 0;
	m_iLightPosY = 0;

	m_InputManager.CreateDevices(m_hWnd, true, true);

  m_SkyBox.RestoreDeviceObjects(m_pd3dDevice,
    "Ch23p1_SkyboxTop.bmp",
    "Ch23p1_SkyboxBottom.bmp",
    "Ch23p1_SkyboxFront.bmp",
    "Ch23p1_SkyboxBack.bmp",
    "Ch23p1_SkyboxLeft.bmp",
    "Ch23p1_SkyboxRight.bmp"
    );

	m_GroundPlane.RestoreDeviceObjects(m_pd3dDevice, "Ch23p1_GroundTexture.bmp", 512.0f, 512.0f, 16);

  m_Water.RestoreDeviceObjects(m_pd3dDevice, 64, "Ch23p1_InPool.jpg");
 	m_pPoolFrame = new CD3DMesh;

  // Create the cubemap, with a format that matches the backbuffer, since
  // we'll be rendering into it
  if( FAILED( hr = m_pd3dDevice->CreateCubeTexture( CUBEMAP_RESOLUTION, 1, D3DUSAGE_RENDERTARGET,
                                                  m_d3dsdBackBuffer.Format,
                                                  D3DPOOL_DEFAULT, &m_pCubeMap ) ) )
    return E_FAIL;

  // We create a separate Z buffer for the cube faces, because user could 
  // resize rendering window so that it is smaller than a cube face. In 
  // this case we cannot use the rendering window Z buffer for cube faces.
  if( FAILED( hr = m_pd3dDevice->CreateDepthStencilSurface(
            CUBEMAP_RESOLUTION, CUBEMAP_RESOLUTION, 
            D3DFMT_D16, D3DMULTISAMPLE_NONE, &m_pCubeFaceZBuffer) ) )
    return E_FAIL;

  TCHAR buf[256];
	DXUtil_ConvertGenericStringToAnsi(buf, "Ch23p1_PoolFrame.x", sizeof(buf));

	if( FAILED( m_pPoolFrame->Create(m_pd3dDevice, buf) ) )
    return E_FAIL;

	m_pPoolFrame->RestoreDeviceObjects(m_pd3dDevice);

  m_pFont->RestoreDeviceObjects();
  m_pFontSmall->RestoreDeviceObjects();
  m_Camera.SetPosition(D3DXVECTOR3(0.0f, 7.0f, -20.0f));
	
	// Set projection matrix
  D3DXMATRIX matProj;
  FLOAT fAspect = ((FLOAT)m_d3dsdBackBuffer.Width) / m_d3dsdBackBuffer.Height;
  D3DXMatrixPerspectiveFovLH( &matProj, D3DX_PI/4, fAspect, 0.1f, 100.0f );
  m_pd3dDevice->SetTransform( D3DTS_PROJECTION, &matProj );

  m_pd3dDevice->SetTextureStageState( 0, D3DTSS_MINFILTER, D3DTEXF_LINEAR );
  m_pd3dDevice->SetTextureStageState( 0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR );
  m_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLOROP,   D3DTOP_MODULATE );
  m_pd3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAOP,   D3DTOP_MODULATE );
  m_pd3dDevice->SetTextureStageState( 1, D3DTSS_COLOROP,   D3DTOP_DISABLE );
  m_pd3dDevice->SetTextureStageState( 1, D3DTSS_ALPHAOP,   D3DTOP_DISABLE );
  m_pd3dDevice->SetTextureStageState( 1, D3DTSS_MINFILTER, D3DTEXF_LINEAR );
  m_pd3dDevice->SetTextureStageState( 1, D3DTSS_MAGFILTER, D3DTEXF_LINEAR );
  

  m_pd3dDevice->SetRenderState( D3DRS_SRCBLEND,  D3DBLEND_ONE );
  m_pd3dDevice->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_ONE );
  m_pd3dDevice->SetRenderState( D3DRS_LIGHTING,  FALSE );
  m_pd3dDevice->SetRenderState( D3DRS_CULLMODE,  D3DCULL_CCW );
  m_pd3dDevice->SetRenderState( D3DRS_SHADEMODE, D3DSHADE_GOURAUD );
  
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc:
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InvalidateDeviceObjects()
{
  m_InputManager.DestroyDevices();
  m_pFont->InvalidateDeviceObjects();
  m_pFontSmall->InvalidateDeviceObjects();
  m_SkyBox.InvalidateDeviceObjects();
	m_GroundPlane.InvalidateDeviceObjects();
  m_Water.InvalidateDeviceObjects();
  if (m_pPoolFrame) {
		m_pPoolFrame->Destroy();
		SAFE_DELETE(m_pPoolFrame);
	}
  
  SAFE_RELEASE( m_pCubeMap );
  SAFE_RELEASE( m_pCubeFaceZBuffer );

  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::DeleteDeviceObjects()
{
  m_pFont->DeleteDeviceObjects();
  m_pFontSmall->DeleteDeviceObjects();
	m_GroundPlane.DeleteDeviceObjects();
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FinalCleanup()
{
  SAFE_DELETE( m_pFont );
  SAFE_DELETE( m_pFontSmall );
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device intialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::ConfirmDevice( D3DCAPS8* pCaps, DWORD dwBehavior,
                                          D3DFORMAT Format )
{
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Message proc function to handle key and menu input
//-----------------------------------------------------------------------------
LRESULT CMyD3DApplication::MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam,
                                    LPARAM lParam )
{
  float f;
  // Pass remaining messages to default handler
  if (uMsg == WM_KEYDOWN) {
    switch(toupper(wParam)) {
      case 'C':
        m_Water.CreateWaterDroplet(RandomNumber(0,64), RandomNumber(0,64), 1, 5);
        break;
      case 'N':
        m_Water.SetHardcodeNormals(!m_Water.GetHardcodeNormals());
        break;
      case 'O':
        f = m_Water.GetEnvBlendFactor();
        f += 0.1f; if (f > 1.0f) f = 1.0f;
        m_Water.SetEnvBlendFactor(f);
        break;
      case 'I':
        f = m_Water.GetEnvBlendFactor();
        f -= 0.1f; if (f < 0.0f) f = 0.0f;
        m_Water.SetEnvBlendFactor(f);
        break;
      case 'L':
        f = m_Water.GetDepth();
        f += 0.1f; if (f > 10.0f) f = 10.0f;
        m_Water.SetDepth(f);
        break;
      case 'K':
        f = m_Water.GetDepth();
        f -= 0.1f; if (f < 0.0f) f = 0.0f;
        m_Water.SetDepth(f);
        break;
 
    }
  }
  return CD3DApplication::MsgProc( hWnd, uMsg, wParam, lParam );
}