/*
#############################################################################

  Ch24p2_ToonShaderSilo.cpp: a program that demonstrates the fire algorithm,
  without any annoying bells and/or whistles.
  
#############################################################################
*/

// include files ////////////////////////////////////////////////////////////
#define STRICT
#include <stdio.h>
#include <math.h>
#include <D3DX8.h>
#include "D3DApp.h"
#include "D3DFile.h"
#include "D3DFont.h"
#include "D3DUtil.h"
#include "DXUtil.h"
#include "D3DHelperFuncs.h"
#include "Ch24p2_resource.h"
#include "Camera.h"
#include "InputManager.h"
#include "SkyBox.h"
#include "Timer.h"
#include "d3dx8mesh.h"

typedef struct
{
	D3DXVECTOR3 pos;
	D3DXVECTOR3 normal;
  D3DCOLOR diffuse;
	float uShade;         //texcoord for stage 0
	float uEdge;         //texcoord for stage 1
} VERTEX_XYZ_NORMAL_1DTEX2;

class CShadeTextureEntry
{
public:
  CShadeTextureEntry(std::string name, std::string filename) {
    m_name = name; m_filename = filename;
  }

  std::string m_name;
  std::string m_filename;
};

#define D3DFVF_XYZ_NORMAL_1DTEX2 \
  D3DFVF_XYZ | D3DFVF_NORMAL | D3DFVF_DIFFUSE | D3DFVF_TEX2 | \
  D3DFVF_TEXCOORDSIZE1( 0 ) | D3DFVF_TEXCOORDSIZE1( 1 )

//-----------------------------------------------------------------------------
// Name: class CMyD3DApplication
// Desc: Application class. The base class (CD3DApplication) provides the 
//       generic functionality needed in all Direct3D samples. CMyD3DApplication 
//       adds functionality specific to this sample program.
//-----------------------------------------------------------------------------
class CMyD3DApplication : public CD3DApplication
{
  // Font for drawing text
  CD3DFont* m_pFont;
  CD3DFont* m_pFontSmall;

  // Scene
  CUserControlledCamera m_Camera;

  // Mouse Input
  CInputManager m_InputManager;

  // teapot
  CD3DMesh *m_pTeapot;
  LPDIRECT3DTEXTURE8 m_pShadeTexture;
  LPDIRECT3DTEXTURE8 m_pSiloTexture;
	
  D3DXMATRIX m_matProj;

  DWORD m_dwShader;
  std::string m_strTextureDesc;

  std::vector<CShadeTextureEntry> m_ShadeTextures;
  std::vector<CShadeTextureEntry> m_SiloTextures;
  int m_iShadeTexture;
  int m_iSiloTexture;
	
protected:
  HRESULT OneTimeSceneInit();
  HRESULT InitDeviceObjects();
  HRESULT RestoreDeviceObjects();
  HRESULT InvalidateDeviceObjects();
  HRESULT DeleteDeviceObjects();
  HRESULT FinalCleanup();
  HRESULT Render();
  HRESULT FrameMove();
  HRESULT ConfirmDevice( D3DCAPS8* pCaps, DWORD dwBehavior, D3DFORMAT Format );
  LRESULT MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam );
  
  HRESULT SetMeshColor(LPDIRECT3DVERTEXBUFFER8 pVB, 
    int iNumVerts, D3DXCOLOR color);
  void SwitchTextures();

  void ProcessInput();

public:
  CMyD3DApplication();
};

//-----------------------------------------------------------------------------
// Name: WinMain()
// Desc: Entry point to the program. Initializes everything, and goes into a
//       message-processing loop. Idle time is used to render the scene.
//-----------------------------------------------------------------------------
INT WINAPI WinMain( HINSTANCE hInst, HINSTANCE, LPSTR, INT )
{
  CMyD3DApplication d3dApp;

  if( FAILED( d3dApp.Create( hInst ) ) )
    return 0;

  return d3dApp.Run();
}

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
CMyD3DApplication::CMyD3DApplication()
{
  m_strWindowTitle    = _T("Ch24p2_ToonShaderSilo");
  m_bUseDepthBuffer   = TRUE;
  m_pFont            = new CD3DFont( _T("Arial"), 12, D3DFONT_BOLD );
  m_pFontSmall       = new CD3DFont( _T("Arial"),  9, D3DFONT_BOLD );
  m_pShadeTexture = NULL;
  m_pSiloTexture = NULL;
}

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::OneTimeSceneInit()
{
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FrameMove()
{
  m_fElapsedTime = 0.001f;
  
	if (m_fElapsedTime < 0.0001f) {
		m_fElapsedTime = 0.0001f;
	}
	
  if (m_fElapsedTime > 0.005f) {
		m_fElapsedTime = 0.005f;
    m_fTime -= m_fElapsedTime - 0.005f;
	}
	
	CTimer::UpdateAll(m_fElapsedTime);
	return S_OK;
}

void CMyD3DApplication::ProcessInput()
{
  float fSpeed = 100.0f * m_fElapsedTime;
  unsigned char m_bKey[256];
  ZeroMemory( m_bKey, 256 );
  GetKeyboardState(m_bKey);
  // Process keyboard input
  if(m_bKey['D'] & 128) m_Camera.AddToVelocity(D3DXVECTOR3(fSpeed, 0.0f, 0.0f)); // Slide Right
  if(m_bKey['A'] & 128) m_Camera.AddToVelocity(D3DXVECTOR3(-fSpeed, 0.0f, 0.0f));// Slide Left
  if(m_bKey['Q'] & 128) m_Camera.AddToVelocity(D3DXVECTOR3(0.0f, fSpeed, 0.0f)); // Slide Up
  if(m_bKey['Z'] & 128) m_Camera.AddToVelocity(D3DXVECTOR3(0.0f, -fSpeed, 0.0f));// Slide Down
  if(m_bKey['W'] & 128) m_Camera.AddToVelocity(D3DXVECTOR3(0.0f, 0.0f, fSpeed)); // Slide Foward
  if(m_bKey['S'] & 128) m_Camera.AddToVelocity(D3DXVECTOR3(0.0f, 0.0f, -fSpeed));// Slide Back
  if(m_bKey['L'] & 128) m_Camera.AddToYawPitchRoll(fSpeed, 0.0f, 0.0f);  // Turn Right
  if(m_bKey['J'] & 128) m_Camera.AddToYawPitchRoll(-fSpeed, 0.0f, 0.0f); // Turn Left
  if(m_bKey['K'] & 128) m_Camera.AddToYawPitchRoll(0.0f, fSpeed, 0.0f); // Turn Down
  if(m_bKey['I'] & 128) m_Camera.AddToYawPitchRoll(0.0f, -fSpeed, 0.0f);// Turn Up

  // mouse look
  DIMOUSESTATE2 dims2;
  m_InputManager.ReadMouse(dims2);

  // play with the divisor constants to change the mouselook sensitivity.
  // I've found that these values most accurately simulate my beloved Q3A setup. :)
  m_Camera.AddToYawPitchRoll(
    (float)dims2.lX*2000.0f*m_fElapsedTime, 
    (float)dims2.lY*2000.0f*m_fElapsedTime, 0.0f);
  
}

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::Render()
{
  
  // these are done here so that you can move the camera around during a freeze
  // frame, ala The Matrix
  D3DXMATRIX matWorld;
  D3DXMatrixIdentity( &matWorld );
  m_pd3dDevice->SetTransform( D3DTS_WORLD, &matWorld );

  ProcessInput();
  m_Camera.Update(m_fElapsedTime);

  m_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, FALSE );
  m_pd3dDevice->SetRenderState( D3DRS_SRCBLEND, D3DBLEND_SRCALPHA );
  m_pd3dDevice->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA );
  
  // Clear the backbuffer
  m_pd3dDevice->Clear( 0L, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER,
                       0xB0B0B0, 1.0f, 0L );
  
  // Now that our fire texture's updated, we can begin rendering the scene
  if( SUCCEEDED( m_pd3dDevice->BeginScene() ) )
  {
    // draw teapot
    D3DXMatrixIdentity(&matWorld);
    
    // Set up the vertex shader constants
    {
      D3DXVECTOR4 vLightDir(0.0f, 0.0f, -1.0f, 0.0f);
	    D3DXVec4Normalize(&vLightDir, &vLightDir);
	    
	    m_pd3dDevice->SetVertexShaderConstant(4, &vLightDir, 1);
	    
	    D3DXMATRIX matWorldViewProj;
      D3DXMATRIX matWorldView, matWorldViewInverse;

      matWorldViewProj = matWorld * m_Camera.GetViewMatrix() * m_matProj;
      matWorldView = matWorld * m_Camera.GetViewMatrix();
      D3DXMatrixInverse(&matWorldViewInverse, NULL, &matWorldView);
      D3DXMatrixTranspose(&matWorldViewProj, &matWorldViewProj);
      
	  	m_pd3dDevice->SetVertexShaderConstant(0, &matWorldViewProj, 4);
	    m_pd3dDevice->SetVertexShaderConstant(5, &matWorldViewInverse, 4);

      D3DXVECTOR4 vConstants(0.5f, 0.0f, 0.0f, 0.0f);
      m_pd3dDevice->SetVertexShaderConstant(9, &vConstants, 1);
    }

    LPDIRECT3DVERTEXBUFFER8 pVB; m_pTeapot->GetLocalMesh()->GetVertexBuffer(&pVB);
    LPDIRECT3DINDEXBUFFER8 pIB; m_pTeapot->GetLocalMesh()->GetIndexBuffer(&pIB);
    
    SetMeshColor(pVB, m_pTeapot->GetLocalMesh()->GetNumVertices(),
      D3DXCOLOR(1.0, 0.0, 0.0, 1.0));

    m_pd3dDevice->SetTransform(D3DTS_WORLD, &matWorld);
    m_pd3dDevice->SetTransform(D3DTS_VIEW, &m_Camera.GetViewMatrix());
    m_pd3dDevice->SetRenderState(D3DRS_ZENABLE, TRUE);
    m_pd3dDevice->SetVertexShader(m_dwShader); // set our vertex shader active


    m_pd3dDevice->SetStreamSource( 0, pVB, sizeof(VERTEX_XYZ_NORMAL_1DTEX2));
    m_pd3dDevice->SetIndices(pIB, 0);

    m_pd3dDevice->SetTextureStageState(0, D3DTSS_ADDRESSU, D3DTADDRESS_CLAMP);
	  m_pd3dDevice->SetTextureStageState(0, D3DTSS_MINFILTER, D3DTEXF_POINT);
	  m_pd3dDevice->SetTextureStageState(0, D3DTSS_MAGFILTER, D3DTEXF_POINT);
	  m_pd3dDevice->SetTextureStageState(0, D3DTSS_MIPFILTER, D3DTEXF_NONE);

    m_pd3dDevice->SetTextureStageState(1, D3DTSS_ADDRESSU, D3DTADDRESS_CLAMP);
	  m_pd3dDevice->SetTextureStageState(1, D3DTSS_MINFILTER, D3DTEXF_POINT);
	  m_pd3dDevice->SetTextureStageState(1, D3DTSS_MAGFILTER, D3DTEXF_POINT);
	  m_pd3dDevice->SetTextureStageState(1, D3DTSS_MIPFILTER, D3DTEXF_NONE);

    
    // pass 1
    {
    
      // set up texture blending modes
      m_pd3dDevice->SetTexture(0, m_pShadeTexture);
	    
	    
	    m_pd3dDevice->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
	    m_pd3dDevice->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_SELECTARG1);
	    
	    m_pd3dDevice->SetTextureStageState(1, D3DTSS_COLORARG1, D3DTA_CURRENT);
	    m_pd3dDevice->SetTextureStageState(1, D3DTSS_COLORARG2, D3DTA_DIFFUSE);	
      m_pd3dDevice->SetTextureStageState(1, D3DTSS_COLOROP, D3DTOP_MODULATE);

      m_pd3dDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 
        m_pTeapot->GetLocalMesh()->GetNumVertices(), 0, 
        m_pTeapot->GetLocalMesh()->GetNumFaces());
    }
    

    // pass 2 - draw siloheutte
    {
      // set up texture blending modes
      m_pd3dDevice->SetTexture(0, m_pSiloTexture);
      m_pd3dDevice->SetTexture(1, m_pSiloTexture);

      m_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
      m_pd3dDevice->SetRenderState( D3DRS_SRCBLEND, D3DBLEND_SRCALPHA );
      m_pd3dDevice->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA );
  
 	    m_pd3dDevice->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_DIFFUSE);
	    m_pd3dDevice->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_SELECTARG1);
	    m_pd3dDevice->SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_DIFFUSE);
	    m_pd3dDevice->SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_SELECTARG1);
	    
      m_pd3dDevice->SetTextureStageState(1, D3DTSS_COLORARG1, D3DTA_TEXTURE);
	    m_pd3dDevice->SetTextureStageState(1, D3DTSS_COLOROP, D3DTOP_SELECTARG1);
	    m_pd3dDevice->SetTextureStageState(1, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);
	    m_pd3dDevice->SetTextureStageState(1, D3DTSS_ALPHAOP, D3DTOP_SELECTARG1);
	    
      m_pd3dDevice->SetTextureStageState(2, D3DTSS_COLOROP, D3DTOP_DISABLE);
      m_pd3dDevice->SetTextureStageState(2, D3DTSS_ALPHAOP, D3DTOP_DISABLE);

      m_pd3dDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 
        m_pTeapot->GetLocalMesh()->GetNumVertices(), 0, 
        m_pTeapot->GetLocalMesh()->GetNumFaces());
    }

    pVB->Release();
    pIB->Release();    
      

    char buf[256];
    _snprintf(buf, 256, "Position: (%.2f, %.2f, %.2f)", 
      (float)m_Camera.GetPosition().x, (float)m_Camera.GetPosition().y, 
      (float)m_Camera.GetPosition().z);

    m_pFontSmall->DrawText( 2,  0,  D3DCOLOR_ARGB(255,255,255,0), buf );
    
    _snprintf(buf, 256, "Shading Texture: %s (T to change)", 
      m_ShadeTextures[m_iShadeTexture].m_name.c_str());

    m_pFontSmall->DrawText( 2,  15,  D3DCOLOR_ARGB(255,255,255,0), buf );
    
    _snprintf(buf, 256, "Silhouette Texture: %s (Y to change)", 
      m_SiloTextures[m_iSiloTexture].m_name.c_str());

    m_pFontSmall->DrawText( 2,  30,  D3DCOLOR_ARGB(255,255,255,0), buf );
    
    // End the scene.
    m_pd3dDevice->EndScene();
  }

  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InitDeviceObjects()
{
  m_pFont->InitDeviceObjects( m_pd3dDevice );
  m_pFontSmall->InitDeviceObjects( m_pd3dDevice );
	
  return S_OK;
}

void CMyD3DApplication::SwitchTextures()
{
  SAFE_RELEASE(m_pShadeTexture);
  SAFE_RELEASE(m_pSiloTexture);

  D3DXCreateTextureFromFile(m_pd3dDevice, 
    m_ShadeTextures[m_iShadeTexture].m_filename.c_str(), &m_pShadeTexture);

  D3DXCreateTextureFromFile(m_pd3dDevice, 
    m_SiloTextures[m_iSiloTexture].m_filename.c_str(), &m_pSiloTexture);
}
HRESULT CMyD3DApplication::SetMeshColor(LPDIRECT3DVERTEXBUFFER8 pVB, 
                                        int iNumVerts, D3DXCOLOR color)
{
  HRESULT hr;

  // lock vertex buffer
  VERTEX_XYZ_NORMAL_1DTEX2 *pVertices;

  if(FAILED(hr = pVB->Lock(0, 
            iNumVerts*sizeof(VERTEX_XYZ_NORMAL_1DTEX2), 
            (BYTE**)&pVertices, 0)))
    return hr;

  // set diffuse color of each vertex
  for (int q=0; q < iNumVerts; q++) {
    pVertices[q].diffuse = color;
  }

  pVB->Unlock();
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::RestoreDeviceObjects()
{
  HRESULT hr;
	m_InputManager.CreateDevices(m_hWnd, true, true);

  m_pTeapot = new CD3DMesh();

  
  if( FAILED( m_pTeapot->Create( m_pd3dDevice, _T("Ch24p2_Teapot.x") ) ) )
    return D3DAPPERR_MEDIANOTFOUND;


  
  //D3DXCreateTorus(m_pd3dDevice, 0.5, 1, 16, 32, &m_pTeapot->m_pSysMemMesh, NULL);

  m_pTeapot->RestoreDeviceObjects( m_pd3dDevice );
  m_pTeapot->SetFVF(m_pd3dDevice, D3DFVF_XYZ_NORMAL_1DTEX2);

  // Create vertex shader
  {
    LPD3DXBUFFER pCode;

    DWORD dwDecl[] =
    {
      D3DVSD_STREAM(0),
      D3DVSD_REG(0, D3DVSDT_FLOAT3),   // position
      D3DVSD_REG(1, D3DVSDT_FLOAT3),   // normal
      D3DVSD_REG(4, D3DVSDT_D3DCOLOR), // diffuse color
      D3DVSD_REG(2, D3DVSDT_FLOAT1),   // u0
      D3DVSD_REG(3, D3DVSDT_FLOAT1),   // u1
      D3DVSD_END()
    };

    // Assemble the vertex shader from the file
    if( FAILED( hr = D3DXAssembleShaderFromFile( "Ch24p2_ToonShaderSilo.vs", 
                                                 0, NULL, &pCode, NULL ) ) )
      return hr;

    // Create the vertex shader
    hr = m_pd3dDevice->CreateVertexShader( dwDecl, 
      (DWORD*)pCode->GetBufferPointer(), &m_dwShader, 0 );
    pCode->Release();
    if( FAILED(hr) ) return hr;
  }


  m_ShadeTextures.clear();
  m_ShadeTextures.push_back(CShadeTextureEntry("3-color", "Ch24p2_ShadeTexture_3.bmp"));
  m_ShadeTextures.push_back(CShadeTextureEntry("16-color", "Ch24p2_ShadeTexture_16.bmp"));
  m_ShadeTextures.push_back(CShadeTextureEntry("Smooth", "Ch24p2_ShadeTexture_256.bmp"));
  m_ShadeTextures.push_back(CShadeTextureEntry("Inverted", "Ch24p2_ShadeTexture_Inverse.bmp"));
  
  m_iShadeTexture = 0;
  
  m_SiloTextures.clear();
  m_SiloTextures.push_back(CShadeTextureEntry("Normal Pen", "Ch24p2_SiloTextureNormal.dds"));
  m_SiloTextures.push_back(CShadeTextureEntry("Thin Pen", "Ch24p2_SiloTextureThin.dds"));
  m_SiloTextures.push_back(CShadeTextureEntry("Pencil", "Ch24p2_SiloTexturePencil.dds"));
  m_SiloTextures.push_back(CShadeTextureEntry("Thick Pen", "Ch24p2_SiloTextureThickPen.dds"));
  m_iSiloTexture = 0;
  
  SwitchTextures();
  
  m_pFont->RestoreDeviceObjects();
  m_pFontSmall->RestoreDeviceObjects();
  m_Camera.SetPosition(D3DXVECTOR3(0.0f, 0.0f, -5.0f));
	
	// Set the world matrix
  D3DXMATRIX matWorld;
  D3DXMatrixIdentity( &matWorld );
  m_pd3dDevice->SetTransform( D3DTS_WORLD, &matWorld );

  // Set projection matrix
  FLOAT fAspect = ((FLOAT)m_d3dsdBackBuffer.Width) / m_d3dsdBackBuffer.Height;
  D3DXMatrixPerspectiveFovLH( &m_matProj, D3DX_PI/4, fAspect, 0.1f, 100.0f );
  m_pd3dDevice->SetTransform( D3DTS_PROJECTION, &m_matProj );

  m_pd3dDevice->SetRenderState( D3DRS_SRCBLEND,  D3DBLEND_ONE );
  m_pd3dDevice->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_ONE );
  m_pd3dDevice->SetRenderState( D3DRS_LIGHTING,  FALSE );
  m_pd3dDevice->SetRenderState( D3DRS_CULLMODE,  D3DCULL_CCW );
  m_pd3dDevice->SetRenderState( D3DRS_SHADEMODE, D3DSHADE_GOURAUD );
  
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc:
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InvalidateDeviceObjects()
{
  m_pd3dDevice->DeleteVertexShader(m_dwShader);
  SAFE_RELEASE(m_pShadeTexture);
  SAFE_RELEASE(m_pSiloTexture);

  m_pTeapot->InvalidateDeviceObjects();
  SAFE_DELETE(m_pTeapot);

  m_InputManager.DestroyDevices();
  m_pFont->InvalidateDeviceObjects();
  m_pFontSmall->InvalidateDeviceObjects();
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::DeleteDeviceObjects()
{
  m_pFont->DeleteDeviceObjects();
  m_pFontSmall->DeleteDeviceObjects();
    
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FinalCleanup()
{
  SAFE_DELETE( m_pFont );
  SAFE_DELETE( m_pFontSmall );
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device intialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::ConfirmDevice( D3DCAPS8* pCaps, DWORD dwBehavior,
                                          D3DFORMAT Format )
{
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Message proc function to handle key and menu input
//-----------------------------------------------------------------------------
LRESULT CMyD3DApplication::MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam,
                                    LPARAM lParam )
{
  if (uMsg == WM_KEYUP && (wParam == 't' || wParam == 'T')) {
    m_iShadeTexture++; if (m_iShadeTexture >= m_ShadeTextures.size()) {
      m_iShadeTexture = 0;
    }
    SwitchTextures();
  }
  if (uMsg == WM_KEYUP && (wParam == 'y' || wParam == 'Y')) {
    m_iSiloTexture++; if (m_iSiloTexture >= m_SiloTextures.size()) {
      m_iSiloTexture = 0;
    }
    SwitchTextures();
  }
 
  // Pass remaining messages to default handler
  return CD3DApplication::MsgProc( hWnd, uMsg, wParam, lParam );
}