/*
#############################################################################

  Ch24p3_ImgManip.cpp: a program that demonstrates the fire algorithm,
  without any annoying bells and/or whistles.
  
#############################################################################
*/

// include files ////////////////////////////////////////////////////////////
#define STRICT
#include <stdio.h>
#include <math.h>
#include <D3DX8.h>
#include "D3DApp.h"
#include "D3DFile.h"
#include "D3DFont.h"
#include "D3DUtil.h"
#include "DXUtil.h"
#include "D3DHelperFuncs.h"
#include "Ch24p3_resource.h"
#include "Camera.h"
#include "InputManager.h"
#include "Timer.h"
#include "d3dx8mesh.h"

typedef struct
{
  D3DXVECTOR3 position; // The position
  D3DCOLOR    color;    // The color
  float       tu0, tv0;   // texcoords 0
  float       tu1, tv1;   // texcoords 1
  float       tu2, tv2;   // texcoords 2
  float       tu3, tv3;   // texcoords 3
} VERTEX_IMGMANIP;

#define D3DFVF_IMGMANIP (D3DFVF_XYZ|D3DFVF_DIFFUSE|D3DFVF_TEX4)

typedef struct
{
  D3DXVECTOR3 position; // The position
  D3DXVECTOR3 normal;   // normal
  D3DCOLOR    color;    // diffuse color
  float       tu, tv;   // texture coordinates
} VERTEX_TEAPOT;

#define D3DFVF_TEAPOT (D3DFVF_XYZ|D3DFVF_NORMAL|D3DFVF_DIFFUSE|D3DFVF_TEX1)


class CPixelShaderEntry
{
public:
  CPixelShaderEntry() { m_dwID = 0; m_pd3dDevice = NULL; }

  virtual ~CPixelShaderEntry() { }
  
  CPixelShaderEntry(LPDIRECT3DDEVICE8 pDev, std::string strName, 
    std::string strFilename)
  {
    m_strName = strName;
    m_strFilename = strFilename;
    m_pd3dDevice = pDev;
    if (FAILED(CreateShader(pDev, strFilename.c_str(), NULL, m_dwID))) {
      throw("Cannot Create Pixel Shader!");
    }
  }

  LPDIRECT3DDEVICE8 m_pd3dDevice;
  std::string m_strName;
  std::string m_strFilename;
  DWORD m_dwID;
};

//-----------------------------------------------------------------------------
// Name: class CMyD3DApplication
// Desc: Application class. The base class (CD3DApplication) provides the 
//       generic functionality needed in all Direct3D samples. CMyD3DApplication 
//       adds functionality specific to this sample program.
//-----------------------------------------------------------------------------
class CMyD3DApplication : public CD3DApplication
{
  // Font for drawing text
  CD3DFont* m_pFont;
  CD3DFont* m_pFontSmall;

  // Scene
  CUserControlledCamera m_Camera;

  // Mouse Input
  CInputManager m_InputManager;

  // teapot
  CD3DMesh *m_pTeapot;
  LPDIRECT3DTEXTURE8 m_texRender;
  LPDIRECT3DSURFACE8 m_surfRenderDepth;
	
  DWORD m_dwTeapotVertexShader;
  DWORD m_dwImageVertexShader;
  
  std::vector<CPixelShaderEntry> m_PixelShaders;
  int m_iCurPixelShader;

  static const int m_iTextureSize;

  float m_fTotalTime;
	
protected:
  HRESULT OneTimeSceneInit();
  HRESULT InitDeviceObjects();
  HRESULT RestoreDeviceObjects();
  HRESULT InvalidateDeviceObjects();
  HRESULT DeleteDeviceObjects();
  HRESULT FinalCleanup();
  HRESULT Render();
  HRESULT FrameMove();
  HRESULT ConfirmDevice( D3DCAPS8* pCaps, DWORD dwBehavior, D3DFORMAT Format );
  LRESULT MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam );

  void ProcessInput();
  HRESULT RenderSceneOntoTexture(D3DXMATRIX matView);

public:
  CMyD3DApplication();
};

const int CMyD3DApplication::m_iTextureSize = 256;

//-----------------------------------------------------------------------------
// Name: WinMain()
// Desc: Entry point to the program. Initializes everything, and goes into a
//       message-processing loop. Idle time is used to render the scene.
//-----------------------------------------------------------------------------
INT WINAPI WinMain( HINSTANCE hInst, HINSTANCE, LPSTR, INT )
{
  CMyD3DApplication d3dApp;

  if( FAILED( d3dApp.Create( hInst ) ) )
    return 0;

  return d3dApp.Run();
}

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
CMyD3DApplication::CMyD3DApplication()
{
  m_strWindowTitle    = _T("Ch24p3_ImgManip");
  m_bUseDepthBuffer   = TRUE;
  m_pFont            = new CD3DFont( _T("Arial"), 12, D3DFONT_BOLD );
  m_pFontSmall       = new CD3DFont( _T("Arial"),  9, D3DFONT_BOLD );
  m_texRender = NULL;
  m_surfRenderDepth = NULL;
  m_iCurPixelShader = 0;
}

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::OneTimeSceneInit()
{
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FrameMove()
{
  m_fElapsedTime = 0.001f;
  
	if (m_fElapsedTime < 0.0001f) {
		m_fElapsedTime = 0.0001f;
	}
	
  if (m_fElapsedTime > 0.005f) {
		m_fElapsedTime = 0.005f;
    m_fTime -= m_fElapsedTime - 0.005f;
	}

  m_fTotalTime += m_fElapsedTime;
	
	CTimer::UpdateAll(m_fElapsedTime);
	return S_OK;
}

void CMyD3DApplication::ProcessInput()
{
  float fSpeed = 600.0f * m_fElapsedTime;
  unsigned char m_bKey[256];
  ZeroMemory( m_bKey, 256 );
  GetKeyboardState(m_bKey);
  // Process keyboard input
  if(m_bKey['D'] & 128) m_Camera.AddToVelocity(D3DXVECTOR3(fSpeed, 0.0f, 0.0f)); // Slide Right
  if(m_bKey['A'] & 128) m_Camera.AddToVelocity(D3DXVECTOR3(-fSpeed, 0.0f, 0.0f));// Slide Left
  if(m_bKey['Q'] & 128) m_Camera.AddToVelocity(D3DXVECTOR3(0.0f, fSpeed, 0.0f)); // Slide Up
  if(m_bKey['Z'] & 128) m_Camera.AddToVelocity(D3DXVECTOR3(0.0f, -fSpeed, 0.0f));// Slide Down
  if(m_bKey['W'] & 128) m_Camera.AddToVelocity(D3DXVECTOR3(0.0f, 0.0f, fSpeed)); // Slide Foward
  if(m_bKey['S'] & 128) m_Camera.AddToVelocity(D3DXVECTOR3(0.0f, 0.0f, -fSpeed));// Slide Back
  if(m_bKey['L'] & 128) m_Camera.AddToYawPitchRoll(fSpeed, 0.0f, 0.0f);  // Turn Right
  if(m_bKey['J'] & 128) m_Camera.AddToYawPitchRoll(-fSpeed, 0.0f, 0.0f); // Turn Left
  if(m_bKey['K'] & 128) m_Camera.AddToYawPitchRoll(0.0f, fSpeed, 0.0f); // Turn Down
  if(m_bKey['I'] & 128) m_Camera.AddToYawPitchRoll(0.0f, -fSpeed, 0.0f);// Turn Up

  // mouse look
  DIMOUSESTATE2 dims2;
  m_InputManager.ReadMouse(dims2);

  // play with the divisor constants to change the mouselook sensitivity.
  // I've found that these values most accurately simulate my beloved Q3A setup. :)
  m_Camera.AddToYawPitchRoll((float)dims2.lX/0.8f, (float)dims2.lY/0.8f, 0.0f);
  
}

HRESULT CMyD3DApplication::RenderSceneOntoTexture(D3DXMATRIX matView)
{
  m_pd3dDevice->Clear( 0L, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER,
                       D3DXCOLOR(0,0,0.5,1), 1.0f, 0L );

  if( SUCCEEDED( m_pd3dDevice->BeginScene() ) )
  {
    
    D3DXMATRIX matProj;
    
    D3DXMatrixPerspectiveFovLH(&matProj, D3DX_PI/2, 1.0f, 0.5f, 100.0f);
    m_pd3dDevice->SetTransform(D3DTS_PROJECTION, &matProj);

    m_pd3dDevice->SetTexture(0, NULL);
    m_pd3dDevice->SetVertexShader(m_dwTeapotVertexShader);

    m_pd3dDevice->SetRenderState(D3DRS_ZENABLE, TRUE);
    m_pd3dDevice->SetRenderState(D3DRS_ALPHABLENDENABLE,  FALSE);

    m_pd3dDevice->SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_SELECTARG1);
	  m_pd3dDevice->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_DIFFUSE);
	  m_pd3dDevice->SetTextureStageState(0, D3DTSS_ALPHAOP,   D3DTOP_SELECTARG1);
    m_pd3dDevice->SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_DIFFUSE);

    m_pd3dDevice->SetTextureStageState(1, D3DTSS_COLOROP,   D3DTOP_DISABLE);
  	m_pd3dDevice->SetTextureStageState(1, D3DTSS_ALPHAOP,   D3DTOP_DISABLE);

    LPDIRECT3DVERTEXBUFFER8 pVB; m_pTeapot->m_pLocalMesh->GetVertexBuffer(&pVB);
    LPDIRECT3DINDEXBUFFER8 pIB; m_pTeapot->m_pLocalMesh->GetIndexBuffer(&pIB);

    for (int q=0; q < 3; q++) {
      D3DXCOLOR colorDiffuse;
      D3DXMATRIX matWorld;
      D3DXMATRIX matInternalRot, matOrbitRot, matTranslate;

      switch(q) {
        case 0: 
          colorDiffuse = D3DXCOLOR(1.0f,0.0f,0.0f,1.0f); 
          D3DXMatrixRotationYawPitchRoll(&matInternalRot, m_fTotalTime*4.0f, m_fTotalTime, 0.0f);
          D3DXMatrixRotationYawPitchRoll(&matOrbitRot, m_fTotalTime*3.0f, 0.0f, m_fTotalTime*3.0f);
          D3DXMatrixIdentity(&matTranslate);
          D3DXMatrixTranslation(&matTranslate, 0.0f, 2.0f, 2.0f);
          break;

        case 1: 
          colorDiffuse = D3DXCOLOR(0.0f,1.0f,0.0f,1.0f); 
          D3DXMatrixRotationYawPitchRoll(&matInternalRot, m_fTotalTime*10.0f, m_fTotalTime, 0.0f);
          D3DXMatrixRotationYawPitchRoll(&matOrbitRot, m_fTotalTime*4.8f, m_fTotalTime, m_fTotalTime*4.8f);
          D3DXMatrixIdentity(&matTranslate);
          D3DXMatrixTranslation(&matTranslate, 0.0f, 1.0f, 1.0f);
          break;
        case 2: 
          colorDiffuse = D3DXCOLOR(0.0f,0.0f,1.0f,1.0f); 
          D3DXMatrixRotationYawPitchRoll(&matInternalRot, m_fTotalTime*6.0f, m_fTotalTime, 0.0f);
          D3DXMatrixRotationYawPitchRoll(&matOrbitRot, m_fTotalTime, 0.0f, m_fTotalTime);
          D3DXMatrixIdentity(&matTranslate);
          D3DXMatrixTranslation(&matTranslate, 0.0f, 3.5f, 3.5f);
          break;
      }

      matWorld = matInternalRot * matTranslate * matOrbitRot;
      m_pd3dDevice->SetTransform(D3DTS_WORLD, &matWorld);
      m_pd3dDevice->SetTransform(D3DTS_VIEW, &matView);

      // Set up the vertex shader constants
      {
        D3DXMATRIX matWorldViewProj;
        D3DXMATRIX matWorldView;

        matWorldViewProj = matWorld * m_Camera.GetViewMatrix() * matProj;
        matWorldView = matWorld * m_Camera.GetViewMatrix();

        D3DXMatrixTranspose(&matWorldViewProj, &matWorldViewProj);
        D3DXVECTOR3 vLightDir(0.0, 0.0, 1.0);
        
	  	  m_pd3dDevice->SetVertexShaderConstant(10, vLightDir, 1);
	      m_pd3dDevice->SetVertexShaderConstant(0, &matWorldViewProj, 4);
	      m_pd3dDevice->SetVertexShaderConstant(11, &matWorldView, 4);
      }

      m_pd3dDevice->SetVertexShaderConstant(5, &colorDiffuse, 1);

      m_pd3dDevice->SetStreamSource( 0, pVB, sizeof(VERTEX_TEAPOT));
      m_pd3dDevice->SetIndices(pIB, 0);
    	  
      m_pd3dDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 
        m_pTeapot->m_pLocalMesh->GetNumVertices(), 0, m_pTeapot->m_pLocalMesh->GetNumFaces());
    }
    
    pVB->Release();
    pIB->Release();

    m_pd3dDevice->EndScene();
  }

  return(S_OK);
}
//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::Render()
{
  // these are done here so that you can move the camera around during a freeze
  // frame, ala The Matrix
  ProcessInput();
  m_Camera.Update(m_fElapsedTime);

  
  // render onto the off-screen texture
  LPDIRECT3DSURFACE8 pDepthSurf;
  m_pd3dDevice->GetDepthStencilSurface(&pDepthSurf);

  LPDIRECT3DSURFACE8 pOldRenderTarget;
  m_pd3dDevice->GetRenderTarget(&pOldRenderTarget);

  LPDIRECT3DSURFACE8 psurfRender;
  m_texRender->GetSurfaceLevel(0, &psurfRender);

  // set new rendering target & clear
  m_pd3dDevice->SetRenderTarget(psurfRender, m_surfRenderDepth);
  
  HRESULT hr;
  if (FAILED(hr = RenderSceneOntoTexture(m_Camera.GetViewMatrix()))) {
    return(hr);
  }

  m_pd3dDevice->SetRenderTarget(pOldRenderTarget, pDepthSurf);
  m_pd3dDevice->SetRenderState(D3DRS_ZENABLE, D3DZB_TRUE);
  
  psurfRender->Release();
  pOldRenderTarget->Release();
  pDepthSurf->Release();
  
  // Clear the backbuffer
  m_pd3dDevice->Clear( 0L, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER,
                       0x000000, 1.0f, 0L );
  
  
  if( SUCCEEDED( m_pd3dDevice->BeginScene() ) )
  {
    D3DXMATRIX projmat;
    D3DXMatrixOrthoLH(&projmat, 1000.0f, 1000.0f, 0.0, 100.0);
    m_pd3dDevice->SetTransform(D3DTS_PROJECTION, &projmat);

    D3DXMATRIX viewmat, worldmat, transmat, scalemat;

    D3DXMatrixIdentity(&viewmat);
    m_pd3dDevice->SetTransform(D3DTS_VIEW, &viewmat);

    D3DXMatrixScaling(&scalemat, 400.0f, 400.0f, 1.0f);
    D3DXMatrixTranslation(&transmat, -250.0f, 0.0f, 0.0f);
    
    worldmat = scalemat * transmat;
    m_pd3dDevice->SetTransform(D3DTS_WORLD, &worldmat);
    
    VERTEX_IMGMANIP pVertices[6];

    // first triangle
    pVertices[0].color = D3DXCOLOR(1.0f,1.0f,1.0f,1.0f);
    pVertices[0].position = D3DXVECTOR3(-0.5f, 0.5f, 0.0f);
    pVertices[0].tu0 = 0.0f; pVertices[0].tv0 = 0.0f;

    pVertices[1].color = D3DXCOLOR(1.0f,1.0f,1.0f,1.0f);
    pVertices[1].position = D3DXVECTOR3(0.5f, 0.5f, 0.0f);
    pVertices[1].tu0 = 1.0f; pVertices[1].tv0 = 0.0f;

    pVertices[2].color = D3DXCOLOR(1.0f,1.0f,1.0f,1.0f);
    pVertices[2].position = D3DXVECTOR3(0.5f, -0.5f, 0.0f);
    pVertices[2].tu0 = 1.0f; pVertices[2].tv0 = 1.0f;

    // second triangle
    pVertices[3].color = D3DXCOLOR(1.0f,1.0f,1.0f,1.0f);
    pVertices[3].position = D3DXVECTOR3(-0.5f, 0.5f, 0.0f);
    pVertices[3].tu0 = 0.0f; pVertices[3].tv0 = 0.0f;

    pVertices[4].color = D3DXCOLOR(1.0f,1.0f,1.0f,1.0f);
    pVertices[4].position = D3DXVECTOR3(0.5f, -0.5f, 0.0f);
    pVertices[4].tu0 = 1.0f; pVertices[4].tv0 = 1.0f;

    pVertices[5].color = D3DXCOLOR(1.0f,1.0f,1.0f,1.0f);
    pVertices[5].position = D3DXVECTOR3(-0.5f, -0.5f, 0.0f);
    pVertices[5].tu0 = 0.0f; pVertices[5].tv0 = 1.0f;

    m_pd3dDevice->SetRenderState(D3DRS_ZENABLE, FALSE);
    m_pd3dDevice->SetRenderState(D3DRS_LIGHTING, FALSE);
    
    m_pd3dDevice->SetTexture(0, m_texRender);
    m_pd3dDevice->SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_SELECTARG1);
	  m_pd3dDevice->SetTextureStageState(0, D3DTSS_ALPHAOP,   D3DTOP_SELECTARG1);
    m_pd3dDevice->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);

	  m_pd3dDevice->SetTextureStageState(1, D3DTSS_COLOROP,   D3DTOP_DISABLE); 
	  
    

    // set normal vertex shader
    m_pd3dDevice->SetVertexShader(D3DFVF_IMGMANIP);

    // set normal pixel shader

    m_pd3dDevice->DrawPrimitiveUP(D3DPT_TRIANGLELIST, 2, pVertices, 
      sizeof(VERTEX_IMGMANIP));
    
    D3DXMatrixScaling(&scalemat, 400.0f, 400.0f, 1.0f);
    D3DXMatrixTranslation(&transmat, 250.0f, 0.0f, 0.0f);
    
    worldmat = scalemat * transmat;
    m_pd3dDevice->SetTransform(D3DTS_WORLD, &worldmat);
    
    m_pd3dDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);

    // set effect pixel shader
    m_pd3dDevice->SetPixelShader(m_PixelShaders[m_iCurPixelShader].m_dwID);
    m_pd3dDevice->SetTexture(0, m_texRender);
    m_pd3dDevice->SetTexture(1, m_texRender);
    m_pd3dDevice->SetTexture(2, m_texRender);
    m_pd3dDevice->SetTexture(3, m_texRender);

    m_pd3dDevice->SetVertexShader(m_dwImageVertexShader);
    // Set up the vertex shader constants
    {
      D3DXMATRIX matWorldViewProj;
    
      matWorldViewProj = worldmat * projmat;
    
      D3DXMatrixTranspose(&matWorldViewProj, &matWorldViewProj);

      float fPerTexelOffset = 0.5 / m_iTextureSize;

      D3DXVECTOR4 vConstants0(-fPerTexelOffset, -fPerTexelOffset, 0.0, 0.0);
      D3DXVECTOR4 vConstants1(-fPerTexelOffset,  fPerTexelOffset, 0.0, 0.0);
      D3DXVECTOR4 vConstants2( fPerTexelOffset, -fPerTexelOffset, 0.0, 0.0);
      D3DXVECTOR4 vConstants3( fPerTexelOffset,  fPerTexelOffset, 0.0, 0.0);
      
	    m_pd3dDevice->SetVertexShaderConstant(0, &matWorldViewProj, 4);
      m_pd3dDevice->SetVertexShaderConstant(5, &vConstants0, 1);
      m_pd3dDevice->SetVertexShaderConstant(6, &vConstants1, 1);
      m_pd3dDevice->SetVertexShaderConstant(7, &vConstants2, 1);
      m_pd3dDevice->SetVertexShaderConstant(8, &vConstants3, 1);
    }

    m_pd3dDevice->DrawPrimitiveUP(D3DPT_TRIANGLELIST, 2, pVertices, 
      sizeof(VERTEX_IMGMANIP));

    m_pd3dDevice->SetPixelShader(NULL);
    
    
    char buf[256];
    _snprintf(buf, 256, "Time: %.5f", m_fTotalTime);
    m_pFontSmall->DrawText( 2,  0,  D3DCOLOR_ARGB(255,255,255,0), buf );

    _snprintf(buf, 256, "On The Left:  Normal Image");
    m_pFontSmall->DrawText( 2,  15,  D3DCOLOR_ARGB(255,0,255,0), buf );
    
    _snprintf(buf, 256, "On The Right: %s ('P' to change)", m_PixelShaders[m_iCurPixelShader].m_strName.c_str());
    m_pFontSmall->DrawText( 2,  30,  D3DCOLOR_ARGB(255,0,255,0), buf );

    // End the scene.
    m_pd3dDevice->EndScene();
  }
  

  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InitDeviceObjects()
{
  m_pFont->InitDeviceObjects( m_pd3dDevice );
  m_pFontSmall->InitDeviceObjects( m_pd3dDevice );
	
  return S_OK;
}


//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::RestoreDeviceObjects()
{
  HRESULT hr;
	m_InputManager.CreateDevices(m_hWnd, true, true);

  m_fTotalTime = 0.0f;
  m_pTeapot = new CD3DMesh();

  if( FAILED( m_pTeapot->Create( m_pd3dDevice, _T("Ch24p3_Teapot.x") ) ) )
    return D3DAPPERR_MEDIANOTFOUND;

  m_pTeapot->RestoreDeviceObjects( m_pd3dDevice );
  m_pTeapot->SetFVF(m_pd3dDevice, D3DFVF_TEAPOT);

  if (FAILED(hr = D3DXCreateTexture(m_pd3dDevice, m_iTextureSize, m_iTextureSize, 
      1, D3DUSAGE_RENDERTARGET, D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, 
      &m_texRender))) throw(hr);

  // create z buffer for off-screen rendering
  if( FAILED( hr = m_pd3dDevice->CreateDepthStencilSurface(
            m_iTextureSize, m_iTextureSize, 
            D3DFMT_D16, D3DMULTISAMPLE_NONE, &m_surfRenderDepth) ) )
    return(hr);


  m_pd3dDevice->SetTextureStageState( 0, D3DTSS_MINFILTER, D3DTEXF_LINEAR );
  m_pd3dDevice->SetTextureStageState( 0, D3DTSS_MAGFILTER, D3DTEXF_LINEAR );
  m_pd3dDevice->SetTextureStageState( 0, D3DTSS_COLOROP,   D3DTOP_MODULATE );
  m_pd3dDevice->SetTextureStageState( 0, D3DTSS_ALPHAOP,   D3DTOP_MODULATE );
  
  // Create teapot vertex shader
  DWORD dwTeapotDecl[] =
  {
    D3DVSD_STREAM(0),
    D3DVSD_REG(0, D3DVSDT_FLOAT3),   // position
    D3DVSD_REG(1, D3DVSDT_FLOAT3),   // normal
    D3DVSD_REG(2, D3DVSDT_D3DCOLOR), // diffuse color
    D3DVSD_REG(3, D3DVSDT_FLOAT2),   // tex coords
    D3DVSD_END()
  };
  CreateShader(m_pd3dDevice, "Ch24p3_Teapot.vs", &dwTeapotDecl[0], m_dwTeapotVertexShader);
  
  DWORD dwImageDecl[] =
  {
    D3DVSD_STREAM(0),
    D3DVSD_REG(0, D3DVSDT_FLOAT3),   // position
    D3DVSD_REG(2, D3DVSDT_D3DCOLOR), // diffuse color
    D3DVSD_REG(3, D3DVSDT_FLOAT2),   // tex coords 0
    D3DVSD_REG(4, D3DVSDT_FLOAT2),   // tex coords 1
    D3DVSD_REG(5, D3DVSDT_FLOAT2),   // tex coords 2
    D3DVSD_REG(6, D3DVSDT_FLOAT2),   // tex coords 3
    D3DVSD_END()
  };

  CreateShader(m_pd3dDevice, "Ch24p3_Image.vs", &dwImageDecl[0], m_dwImageVertexShader);

  m_PixelShaders.push_back(CPixelShaderEntry(m_pd3dDevice, "Normal", "Ch24p3_Normal.ps"));
  m_PixelShaders.push_back(CPixelShaderEntry(m_pd3dDevice, "Black And White", "Ch24p3_BlackAndWhite.ps"));
  m_PixelShaders.push_back(CPixelShaderEntry(m_pd3dDevice, "Better Black And White", "Ch24p3_BetterBlackAndWhite.ps"));
  m_PixelShaders.push_back(CPixelShaderEntry(m_pd3dDevice, "Funky Monochrome", "Ch24p3_Monochrome.ps"));
  m_PixelShaders.push_back(CPixelShaderEntry(m_pd3dDevice, "Blur", "Ch24p3_Blur.ps"));
  m_PixelShaders.push_back(CPixelShaderEntry(m_pd3dDevice, "Sharpen", "Ch24p3_Sharpen.ps"));
  m_PixelShaders.push_back(CPixelShaderEntry(m_pd3dDevice, "Inverse", "Ch24p3_Inverse.ps"));

  m_pFont->RestoreDeviceObjects();
  m_pFontSmall->RestoreDeviceObjects();
  m_Camera.SetPosition(D3DXVECTOR3(0.0f, 0.0f, -5.0f));
	
  m_pd3dDevice->SetRenderState( D3DRS_SRCBLEND,  D3DBLEND_ONE );
  m_pd3dDevice->SetRenderState( D3DRS_DESTBLEND, D3DBLEND_ONE );
  m_pd3dDevice->SetRenderState( D3DRS_LIGHTING,  FALSE );
  m_pd3dDevice->SetRenderState( D3DRS_CULLMODE,  D3DCULL_CCW );
  m_pd3dDevice->SetRenderState( D3DRS_SHADEMODE, D3DSHADE_GOURAUD );
  
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc:
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InvalidateDeviceObjects()
{
  m_pd3dDevice->DeleteVertexShader(m_dwTeapotVertexShader);
  m_pd3dDevice->DeleteVertexShader(m_dwImageVertexShader);
  SAFE_RELEASE(m_texRender);
  SAFE_RELEASE(m_surfRenderDepth);

  // yeah, this isn't good OOD... ideally this deletion would occur in the
  // destructor, 
  for (int q=0; q < m_PixelShaders.size(); q++) {
    m_pd3dDevice->DeletePixelShader(m_PixelShaders[q].m_dwID);
  }
  m_PixelShaders.clear();
  
  m_pTeapot->InvalidateDeviceObjects();
  SAFE_DELETE(m_pTeapot);

  m_InputManager.DestroyDevices();
  m_pFont->InvalidateDeviceObjects();
  m_pFontSmall->InvalidateDeviceObjects();
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::DeleteDeviceObjects()
{
  m_pFont->DeleteDeviceObjects();
  m_pFontSmall->DeleteDeviceObjects();
    
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FinalCleanup()
{
  SAFE_DELETE( m_pFont );
  SAFE_DELETE( m_pFontSmall );
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device intialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::ConfirmDevice( D3DCAPS8* pCaps, DWORD dwBehavior,
                                          D3DFORMAT Format )
{
  return S_OK;
}

//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Message proc function to handle key and menu input
//-----------------------------------------------------------------------------
LRESULT CMyD3DApplication::MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam,
                                    LPARAM lParam )
{
  if (uMsg == WM_KEYUP && (wParam == 'p' || wParam == 'P')) {
    m_iCurPixelShader++;
    if (m_iCurPixelShader >= m_PixelShaders.size()) m_iCurPixelShader = 0;
  }
  // Pass remaining messages to default handler
  return CD3DApplication::MsgProc( hWnd, uMsg, wParam, lParam );
}